# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
import textwrap
import time
import unittest
import uuid
import zipfile
from abc import abstractmethod
from io import BytesIO
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlparse

try:
    from flask import Flask
    from flask.testing import FlaskClient
    from werkzeug.test import TestResponse
except ImportError:
    # We need to let pytest collect tests form this file even if the server
    # extra wasn't installed. We'll then skip them all.
    pass

from toil.test import ToilTest, needs_aws_s3, needs_celery_broker, needs_server

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


@needs_server
class ToilServerUtilsTest(ToilTest):
    """
    Tests for the utility functions used by the Toil server.
    """

    def test_workflow_canceling_recovery(self):
        """
        Make sure that a workflow in CANCELING state will be recovered to a
        terminal state eventually even if the workflow runner Celery task goes
        away without flipping the state.
        """

        from toil.server.utils import (
            MemoryStateStore,
            WorkflowStateMachine,
            WorkflowStateStore,
        )

        store = WorkflowStateStore(MemoryStateStore(), "test-workflow")

        state_machine = WorkflowStateMachine(store)

        # Cancel a workflow
        state_machine.send_cancel()
        # Make sure it worked.
        self.assertEqual(state_machine.get_current_state(), "CANCELING")

        # Back-date the time of cancelation to something really old
        store.set("cancel_time", "2011-11-04 00:05:23.283")

        # Make sure it is now CANCELED due to timeout
        self.assertEqual(state_machine.get_current_state(), "CANCELED")


class hidden:
    # Hide abstract tests from the test loader

    @needs_server
    class AbstractStateStoreTest(ToilTest):
        """
        Basic tests for state stores.
        """

        from toil.server.utils import AbstractStateStore

        @abstractmethod
        def get_state_store(self) -> AbstractStateStore:
            """
            Make a state store to test, on a single fixed URL.
            """

            raise NotImplementedError()

        def test_state_store(self) -> None:
            """
            Make sure that the state store under test can store and load keys.
            """

            store = self.get_state_store()

            # Should start None
            self.assertEqual(store.get("id1", "key1"), None)

            # Should hold a value
            store.set("id1", "key1", "value1")
            self.assertEqual(store.get("id1", "key1"), "value1")

            # Should distinguish by ID and key
            self.assertEqual(store.get("id2", "key1"), None)
            self.assertEqual(store.get("id1", "key2"), None)

            store.set("id2", "key1", "value2")
            store.set("id1", "key2", "value3")
            self.assertEqual(store.get("id1", "key1"), "value1")
            self.assertEqual(store.get("id2", "key1"), "value2")
            self.assertEqual(store.get("id1", "key2"), "value3")

            # Should allow replacement
            store.set("id1", "key1", "value4")
            self.assertEqual(store.get("id1", "key1"), "value4")
            self.assertEqual(store.get("id2", "key1"), "value2")
            self.assertEqual(store.get("id1", "key2"), "value3")

            # Should show up in another state store
            store2 = self.get_state_store()
            self.assertEqual(store2.get("id1", "key1"), "value4")
            self.assertEqual(store2.get("id2", "key1"), "value2")
            self.assertEqual(store2.get("id1", "key2"), "value3")

            # Should allow clearing
            store.set("id1", "key1", None)
            self.assertEqual(store.get("id1", "key1"), None)
            self.assertEqual(store.get("id2", "key1"), "value2")
            self.assertEqual(store.get("id1", "key2"), "value3")

            store.set("id2", "key1", None)
            store.set("id1", "key2", None)
            self.assertEqual(store.get("id1", "key1"), None)
            self.assertEqual(store.get("id2", "key1"), None)
            self.assertEqual(store.get("id1", "key2"), None)


class FileStateStoreTest(hidden.AbstractStateStoreTest):
    """
    Test file-based state storage.
    """

    from toil.server.utils import AbstractStateStore

    def setUp(self) -> None:
        super().setUp()
        self.state_store_dir = self._createTempDir()

    def get_state_store(self) -> AbstractStateStore:
        """
        Make a state store to test, on a single fixed local path.
        """

        from toil.server.utils import FileStateStore

        return FileStateStore(self.state_store_dir)


class FileStateStoreURLTest(hidden.AbstractStateStoreTest):
    """
    Test file-based state storage using URLs instead of local paths.
    """

    from toil.server.utils import AbstractStateStore

    def setUp(self) -> None:
        super().setUp()
        self.state_store_dir = "file://" + self._createTempDir()

    def get_state_store(self) -> AbstractStateStore:
        """
        Make a state store to test, on a single fixed URL.
        """

        from toil.server.utils import FileStateStore

        return FileStateStore(self.state_store_dir)


@needs_aws_s3
class BucketUsingTest(ToilTest):
    """
    Base class for tests that need a bucket.
    """

    if TYPE_CHECKING:
        from mypy_boto3_s3 import S3ServiceResource
        from mypy_boto3_s3.service_resource import Bucket

    region: Optional[str]
    s3_resource: Optional["S3ServiceResource"]
    bucket: Optional["Bucket"]
    bucket_name: Optional[str]

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up the class with a single pre-existing AWS bucket for all tests.
        """
        super().setUpClass()

        from toil.lib.aws import get_current_aws_region, session
        from toil.lib.aws.utils import create_s3_bucket

        cls.region = get_current_aws_region()
        cls.s3_resource = session.resource("s3", region_name=cls.region)

        cls.bucket_name = f"toil-test-{uuid.uuid4()}"
        cls.bucket = create_s3_bucket(cls.s3_resource, cls.bucket_name, cls.region)
        cls.bucket.wait_until_exists()

    @classmethod
    def tearDownClass(cls) -> None:
        from toil.lib.aws.utils import delete_s3_bucket

        if cls.bucket_name:
            delete_s3_bucket(cls.s3_resource, cls.bucket_name, cls.region)
        super().tearDownClass()


class AWSStateStoreTest(hidden.AbstractStateStoreTest, BucketUsingTest):
    """Test AWS-based state storage."""

    from toil.server.utils import AbstractStateStore

    bucket_path = "prefix/of/keys"

    def get_state_store(self) -> AbstractStateStore:
        """
        Make a state store to test, on a single fixed URL.
        """

        from toil.server.utils import S3StateStore

        return S3StateStore("s3://" + self.bucket_name + "/" + self.bucket_path)

    def test_state_store_paths(self) -> None:
        """
        Make sure that the S3 state store puts things in the right places.

        We don't *really* care about the exact internal structure, but we do
        care about actually being under the path we are supposed to use.
        """

        from toil.lib.aws.utils import get_object_for_url

        store = self.get_state_store()

        # Should hold a value
        store.set("testid", "testkey", "testvalue")
        self.assertEqual(store.get("testid", "testkey"), "testvalue")

        expected_url = urlparse(
            "s3://"
            + self.bucket_name
            + "/"
            + os.path.join(self.bucket_path, "testid", "testkey")
        )

        obj = get_object_for_url(expected_url, True)
        self.assertEqual(obj.content_length, len("testvalue"))


@needs_server
class AbstractToilWESServerTest(ToilTest):
    """
    Class for server tests that provides a self.app in testing mode.
    """

    def __init__(self, *args, **kwargs):
        """
        Set up default settings for test classes based on this one.
        """
        super().__init__(*args, **kwargs)

        # Default to the local testing task runner instead of Celery for when
        # we run workflows.
        self._server_args = ["--bypass_celery"]

    def setUp(self) -> None:
        super().setUp()
        self.temp_dir = self._createTempDir()

        from toil.server.app import create_app, parser_with_server_options

        parser = parser_with_server_options()
        args = parser.parse_args(
            self._server_args + ["--work_dir", os.path.join(self.temp_dir, "workflows")]
        )

        # Make the FlaskApp
        server_app = create_app(args)

        # Fish out the actual Flask
        self.app: Flask = server_app.app
        self.app.testing = True

        self.example_cwl = textwrap.dedent(
            """
            cwlVersion: v1.0
            class: CommandLineTool
            baseCommand: echo
            stdout: output.txt
            inputs:
              message:
                type: string
                inputBinding:
                  position: 1
            outputs:
              output:
                type: stdout
            """
        )

        self.slow_cwl = textwrap.dedent(
            """
            cwlVersion: v1.0
            class: CommandLineTool
            baseCommand: sleep
            stdout: output.txt
            inputs:
              delay:
                type: string
                inputBinding:
                  position: 1
            outputs:
              output:
                type: stdout
            """
        )

    def tearDown(self) -> None:
        super().tearDown()

    def _fetch_run_log(self, client: "FlaskClient", run_id: str) -> "TestResponse":
        """
        Fetch the run log for a given workflow.
        """
        rv = client.get(f"/ga4gh/wes/v1/runs/{run_id}")
        self.assertEqual(rv.status_code, 200)
        self.assertTrue(rv.is_json)
        return rv

    def _check_successful_log(self, client: "FlaskClient", run_id: str) -> None:
        """
        Make sure the run log for the given run is generated correctly.
        The workflow should succeed, it should have some tasks, and they should have all succeeded.
        """
        rv = self._fetch_run_log(client, run_id)
        logger.debug("Log info: %s", rv.json)
        run_log = rv.json.get("run_log")
        self.assertEqual(type(run_log), dict)
        if "exit_code" in run_log:
            # The workflow succeeded if it has an exit code
            self.assertEqual(run_log["exit_code"], 0)
        # The workflow is complete
        self.assertEqual(rv.json.get("state"), "COMPLETE")
        task_logs = rv.json.get("task_logs")
        # There are tasks reported
        self.assertEqual(type(task_logs), list)
        self.assertGreater(len(task_logs), 0)
        for task_log in task_logs:
            # All the tasks succeeded
            self.assertEqual(type(task_log), dict)
            self.assertEqual(task_log.get("exit_code"), 0)

    def _report_log(self, client: "FlaskClient", run_id: str) -> None:
        """Report the log for the given workflow run."""
        rv = self._fetch_run_log(client, run_id)
        logger.debug(f"Report log response: {rv.json}")
        run_log = rv.json.get("run_log")
        self.assertEqual(type(run_log), dict)
        self.assertEqual(type(run_log.get("stdout")), str)
        self.assertEqual(type(run_log.get("stderr")), str)
        self._report_absolute_url(client, run_log.get("stdout"))
        self._report_absolute_url(client, run_log.get("stderr"))

    def _report_absolute_url(self, client: "FlaskClient", url: str):
        """
        Take a URL that has a hostname and port, relativize it to the test
        Flask client, and get its contents and log them.
        """

        breakpoint = url.find("/toil/")
        if breakpoint == -1:
            raise RuntimeError("Could not find place to break URL: " + url)
        url = url[breakpoint:]
        logger.info("Fetch %s", url)
        rv = client.get(url)
        self.assertEqual(rv.status_code, 200)
        logger.info("Got %s:\n%s", url, rv.data.decode("utf-8"))

    def _start_slow_workflow(self, client: "FlaskClient") -> str:
        """
        Start a slow workflow and return its ID.
        """
        rv = client.post(
            "/ga4gh/wes/v1/runs",
            data={
                "workflow_url": "slow.cwl",
                "workflow_type": "CWL",
                "workflow_type_version": "v1.0",
                "workflow_params": json.dumps({"delay": "5"}),
                "workflow_attachment": [
                    (BytesIO(self.slow_cwl.encode()), "slow.cwl"),
                ],
            },
        )
        # workflow is submitted successfully
        self.assertEqual(rv.status_code, 200)
        self.assertTrue(rv.is_json)
        run_id = rv.json.get("run_id")
        self.assertIsNotNone(run_id)

        return run_id

    def _poll_status(self, client: "FlaskClient", run_id: str) -> str:
        """
        Get the status of the given workflow.
        """

        rv = client.get(f"/ga4gh/wes/v1/runs/{run_id}/status")
        self.assertEqual(rv.status_code, 200)
        self.assertTrue(rv.is_json)
        self.assertIn("run_id", rv.json)
        self.assertEqual(rv.json.get("run_id"), run_id)
        self.assertIn("state", rv.json)
        state = rv.json.get("state")
        self.assertIn(
            state,
            [
                "UNKNOWN",
                "QUEUED",
                "INITIALIZING",
                "RUNNING",
                "PAUSED",
                "COMPLETE",
                "EXECUTOR_ERROR",
                "SYSTEM_ERROR",
                "CANCELED",
                "CANCELING",
            ],
        )
        return state

    def _cancel_workflow(self, client: "FlaskClient", run_id: str) -> None:
        rv = client.post(f"/ga4gh/wes/v1/runs/{run_id}/cancel")
        self.assertEqual(rv.status_code, 200)

    def _wait_for_status(
        self, client: "FlaskClient", run_id: str, target_status: str
    ) -> None:
        """
        Wait for the given workflow run to reach the given state. If it reaches
        a different terminal state, raise an exception.
        """

        while True:
            state = self._poll_status(client, run_id)
            if state == target_status:
                # We're done!
                logger.info("Workflow reached state %s", state)
                return
            if state in ["EXECUTOR_ERROR", "SYSTEM_ERROR", "CANCELED", "COMPLETE"]:
                logger.error("Workflow run reached unexpected terminal state %s", state)
                self._report_log(client, run_id)
                raise RuntimeError("Workflow is in unexpected state " + state)
            logger.info("Waiting on workflow in state %s", state)
            time.sleep(2)

    def _wait_for_success(self, client: "FlaskClient", run_id: str) -> None:
        """
        Wait for the given workflow run to succeed. If it fails, raise an exception.
        """
        self._wait_for_status(client, run_id, "COMPLETE")
        self._check_successful_log(client, run_id)


class ToilWESServerBenchTest(AbstractToilWESServerTest):
    """
    Tests for Toil's Workflow Execution Service API that don't run workflows.
    """

    def test_home(self) -> None:
        """Test the homepage endpoint."""
        with self.app.test_client() as client:
            rv = client.get("/")
        self.assertEqual(rv.status_code, 302)

    def test_health(self) -> None:
        """Test the health check endpoint."""
        with self.app.test_client() as client:
            rv = client.get("/engine/v1/status")
        self.assertEqual(rv.status_code, 200)

    def test_get_service_info(self) -> None:
        """Test the GET /service-info endpoint."""
        with self.app.test_client() as client:
            rv = client.get("/ga4gh/wes/v1/service-info")
        self.assertEqual(rv.status_code, 200)
        service_info = json.loads(rv.data)

        self.assertIn("version", service_info)
        self.assertIn("workflow_type_versions", service_info)
        self.assertIn("supported_wes_versions", service_info)
        self.assertIn("supported_filesystem_protocols", service_info)
        self.assertIn("workflow_engine_versions", service_info)
        engine_versions = service_info["workflow_engine_versions"]
        self.assertIn("toil", engine_versions)
        self.assertEqual(type(engine_versions["toil"]), str)
        self.assertIn("default_workflow_engine_parameters", service_info)
        self.assertIn("system_state_counts", service_info)
        self.assertIn("tags", service_info)


class ToilWESServerWorkflowTest(AbstractToilWESServerTest):
    """
    Tests of the WES server running workflows.
    """

    def run_zip_workflow(
        self, zip_path: str, include_message: bool = True, include_params: bool = True
    ) -> None:
        """
        We have several zip file tests; this submits a zip file and makes sure it ran OK.

        If include_message is set to False, don't send a "message" argument in  workflow_params.
        If include_params is also set to False, don't send workflow_params at all.
        """
        self.assertTrue(os.path.exists(zip_path))

        # Set up what we will POST to start the workflow.
        post_data = {
            "workflow_url": "file://" + zip_path,
            "workflow_type": "CWL",
            "workflow_type_version": "v1.0",
        }
        if include_params or include_message:
            # We need workflow_params too
            post_data["workflow_params"] = json.dumps(
                {"message": "Hello, world!"} if include_message else {}
            )
        with self.app.test_client() as client:
            rv = client.post("/ga4gh/wes/v1/runs", data=post_data)
            # workflow is submitted successfully
            self.assertEqual(rv.status_code, 200)
            self.assertTrue(rv.is_json)
            run_id = rv.json.get("run_id")
            self.assertIsNotNone(run_id)

            # Check status
            self._wait_for_success(client, run_id)

            # TODO: Make sure that the correct message was output!

    def test_run_workflow_relative_url_no_attachments_fails(self) -> None:
        """Test run example CWL workflow from relative workflow URL but with no attachments."""
        with self.app.test_client() as client:
            rv = client.post(
                "/ga4gh/wes/v1/runs",
                data={
                    "workflow_url": "example.cwl",
                    "workflow_type": "CWL",
                    "workflow_type_version": "v1.0",
                    "workflow_params": "{}",
                },
            )
            self.assertEqual(rv.status_code, 400)
            self.assertTrue(rv.is_json)
            self.assertEqual(
                rv.json.get("msg"),
                "Relative 'workflow_url' but missing 'workflow_attachment'",
            )

    def test_run_workflow_relative_url(self) -> None:
        """Test run example CWL workflow from relative workflow URL."""
        with self.app.test_client() as client:
            rv = client.post(
                "/ga4gh/wes/v1/runs",
                data={
                    "workflow_url": "example.cwl",
                    "workflow_type": "CWL",
                    "workflow_type_version": "v1.0",
                    "workflow_params": json.dumps({"message": "Hello, world!"}),
                    "workflow_attachment": [
                        (BytesIO(self.example_cwl.encode()), "example.cwl"),
                    ],
                },
            )
            # workflow is submitted successfully
            self.assertEqual(rv.status_code, 200)
            self.assertTrue(rv.is_json)
            run_id = rv.json.get("run_id")
            self.assertIsNotNone(run_id)

            # Check status
            self._wait_for_success(client, run_id)

    def test_run_workflow_https_url(self) -> None:
        """Test run example CWL workflow from the Internet."""
        with self.app.test_client() as client:
            rv = client.post(
                "/ga4gh/wes/v1/runs",
                data={
                    "workflow_url": "https://raw.githubusercontent.com/DataBiosphere/toil/4cb5bb3871ac21a9793f638b83775926ed94a226/src/toil/test/cwl/echo.cwl",
                    "workflow_type": "CWL",
                    "workflow_type_version": "v1.2",
                    "workflow_params": json.dumps({"message": "Hello, world!"}),
                },
            )
            # workflow is submitted successfully
            self.assertEqual(rv.status_code, 200)
            self.assertTrue(rv.is_json)
            run_id = rv.json.get("run_id")
            self.assertIsNotNone(run_id)

            # Check status
            self._wait_for_success(client, run_id)

    def test_run_workflow_single_file_zip(self) -> None:
        """Test run example CWL workflow from single-file ZIP."""
        workdir = self._createTempDir()
        zip_path = os.path.abspath(os.path.join(workdir, "workflow.zip"))
        with zipfile.ZipFile(zip_path, "w") as zip_file:
            zip_file.writestr("example.cwl", self.example_cwl)
        self.run_zip_workflow(zip_path)

    def test_run_workflow_multi_file_zip(self) -> None:
        """Test run example CWL workflow from multi-file ZIP."""
        workdir = self._createTempDir()
        zip_path = os.path.abspath(os.path.join(workdir, "workflow.zip"))
        with zipfile.ZipFile(zip_path, "w") as zip_file:
            zip_file.writestr("main.cwl", self.example_cwl)
            zip_file.writestr("distraction.cwl", "Don't mind me")
        self.run_zip_workflow(zip_path)

    def test_run_workflow_manifest_zip(self) -> None:
        """Test run example CWL workflow from ZIP with manifest."""
        workdir = self._createTempDir()
        zip_path = os.path.abspath(os.path.join(workdir, "workflow.zip"))
        with zipfile.ZipFile(zip_path, "w") as zip_file:
            zip_file.writestr("actual.cwl", self.example_cwl)
            zip_file.writestr("distraction.cwl", self.example_cwl)
            zip_file.writestr(
                "MANIFEST.json", json.dumps({"mainWorkflowURL": "actual.cwl"})
            )
        self.run_zip_workflow(zip_path)

    def test_run_workflow_inputs_zip(self) -> None:
        """Test run example CWL workflow from ZIP without manifest but with inputs."""
        workdir = self._createTempDir()
        zip_path = os.path.abspath(os.path.join(workdir, "workflow.zip"))
        with zipfile.ZipFile(zip_path, "w") as zip_file:
            zip_file.writestr("main.cwl", self.example_cwl)
            zip_file.writestr("inputs.json", json.dumps({"message": "Hello, world!"}))
        self.run_zip_workflow(zip_path, include_message=False)

    def test_run_workflow_manifest_and_inputs_zip(self) -> None:
        """Test run example CWL workflow from ZIP with manifest and inputs."""
        workdir = self._createTempDir()
        zip_path = os.path.abspath(os.path.join(workdir, "workflow.zip"))
        with zipfile.ZipFile(zip_path, "w") as zip_file:
            zip_file.writestr("actual.cwl", self.example_cwl)
            zip_file.writestr("data.json", json.dumps({"message": "Hello, world!"}))
            zip_file.writestr(
                "MANIFEST.json",
                json.dumps(
                    {"mainWorkflowURL": "actual.cwl", "inputFileURLs": ["data.json"]}
                ),
            )
        self.run_zip_workflow(zip_path, include_message=False)

    def test_run_workflow_no_params_zip(self) -> None:
        """Test run example CWL workflow from ZIP without workflow_params."""
        workdir = self._createTempDir()
        zip_path = os.path.abspath(os.path.join(workdir, "workflow.zip"))
        with zipfile.ZipFile(zip_path, "w") as zip_file:
            zip_file.writestr("main.cwl", self.example_cwl)
            zip_file.writestr("inputs.json", json.dumps({"message": "Hello, world!"}))
        # Don't even bother sending workflow_params
        self.run_zip_workflow(zip_path, include_message=False, include_params=False)

    def test_run_and_cancel_workflows(self) -> None:
        """
        Run two workflows, cancel one of them, and make sure they all exist.
        """
        with self.app.test_client() as client:
            # Start 1 workflow
            run1 = self._start_slow_workflow(client)
            time.sleep(2)
            status1 = self._poll_status(client, run1)
            self.assertIn(status1, ["QUEUED", "INITIALIZING", "RUNNING"])

            # Start another workflow
            run2 = self._start_slow_workflow(client)
            self.assertNotEqual(run1, run2)
            time.sleep(2)
            status2 = self._poll_status(client, run2)
            self.assertIn(status2, ["QUEUED", "INITIALIZING", "RUNNING"])

            # Cancel the second one
            cancel_sent = time.time()
            self._cancel_workflow(client, run2)
            time.sleep(1)
            status2 = self._poll_status(client, run2)
            self.assertIn(status2, ["CANCELING", "CANCELED"])

            # Make sure the first one still exists too
            status1 = self._poll_status(client, run1)
            self.assertIn(status1, ["QUEUED", "INITIALIZING", "RUNNING"])

            self._wait_for_status(client, run2, "CANCELED")
            cancel_complete = time.time()
            self._wait_for_success(client, run1)

            # Make sure the cancellation was relatively prompt and we didn't
            # have to go through any of the timeout codepaths
            cancel_seconds = cancel_complete - cancel_sent
            logger.info("Cancellation took %s seconds to complete", cancel_seconds)
            from toil.server.wes.tasks import WAIT_FOR_DEATH_TIMEOUT

            self.assertLess(cancel_seconds, WAIT_FOR_DEATH_TIMEOUT)


@needs_celery_broker
class ToilWESServerCeleryWorkflowTest(ToilWESServerWorkflowTest):
    """
    End-to-end workflow-running tests against Celery.
    """

    def __init__(self, *args, **kwargs):
        """
        Set the task runner back to Celery.
        """
        super().__init__(*args, **kwargs)
        self._server_args = []


@needs_celery_broker
class ToilWESServerCeleryS3StateWorkflowTest(
    ToilWESServerWorkflowTest, BucketUsingTest
):
    """
    Test the server with Celery and state stored in S3.
    """

    def setUp(self) -> None:
        # Overwrite server args from __init__. The bucket name isn't available when __init__ runs.
        self._server_args = ["--state_store", "s3://" + self.bucket_name + "/state"]
        super().setUp()


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from toil.job import Job
from toil.realtimeLogger import RealtimeLogger
from toil.test import ToilTest


class RealtimeLoggerTest(ToilTest):
    def testRealtimeLogger(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.realTimeLogging = True
        options.logLevel = "INFO"

        detector = MessageDetector()

        # Set up a log message detector to the root logger
        logging.getLogger().addHandler(detector)

        # I believe coloredlogs replaces handlers with its own when doing handler formatting, preserving only filters
        # https://github.com/xolox/python-coloredlogs/blob/65bdfe976ac0bf81e8c0bd9a98242b9d666b2859/coloredlogs/__init__.py#L453-L459
        options.colored_logs = False
        Job.Runner.startToil(LogTest(), options)

        # We need the message we're supposed to see
        self.assertTrue(detector.detected)
        # But not the message that shouldn't be logged.
        self.assertFalse(detector.overLogged)


class MessageDetector(logging.StreamHandler):
    """
    Detect the secret message and set a flag.
    """

    def __init__(self):
        self.detected = False  # Have we seen the message we want?
        self.overLogged = False  # Have we seen the message we don't want?

        super().__init__()

    def emit(self, record):
        if record.msg == "This should be logged at info level":
            self.detected = True
        if record.msg == "This should be logged at debug level":
            self.overLogged = True


class LogTest(Job):
    def __init__(self):
        Job.__init__(self, memory=100000, cores=1, disk="3G")

    def run(self, fileStore):
        RealtimeLogger.info("This should be logged at info level")
        RealtimeLogger.debug("This should be logged at debug level")



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp

if __name__ == "__main__":
    # A is a job with children and follow-ons, for example:
    A = Job()
    A.addChild(Job())
    A.addFollowOn(Job())

    # B is a job which needs to run after A and its successors
    B = Job()

    # The way to do this without encapsulation is to make a parent of A, Ap, and make B a follow-on of Ap.
    Ap = Job()
    Ap.addChild(A)
    Ap.addFollowOn(B)

    jobstore: str = mkdtemp("tutorial_encapsulations")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "INFO"
    options.clean = "always"

    with Toil(options) as toil:
        print(toil.start(Ap))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os

good_spot = os.path.expanduser('~/.ssh')
if not os.path.exists(good_spot):
    os.mkdir(good_spot)


def format_private_key(private_key):
    private_key_header = '-----BEGIN RSA PRIVATE KEY-----'
    private_key_footer = '-----END RSA PRIVATE KEY-----'

    assert private_key.startswith(private_key_header)
    private_key_body = private_key[len(private_key_header):]
    assert private_key_body.endswith(private_key_footer)
    private_key_body = private_key_body[:-len(private_key_footer)]
    return private_key_header + private_key_body.replace(' ', '\n') + private_key_footer


with open(os.environ['GITLAB_SECRET_FILE_SSH_KEYS'], 'r') as keys_json_file:
    keys = json.loads(keys_json_file.read())
with open(os.path.join(good_spot, 'id_rsa.pub'), 'w') as f:
    f.write(keys['public'])
with open(os.path.join(good_spot, 'id_rsa'), 'w') as f:
    f.write(format_private_key(keys['private']))



# Copyright (C) 2015-2024 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from mypy_boto3_s3.type_defs import ListMultipartUploadsOutputTypeDef

from toil.lib.aws import AWSServerErrors, session
from toil.lib.retry import retry

logger = logging.getLogger(__name__)


@retry(errors=[AWSServerErrors])
def list_multipart_uploads(
    bucket: str, region: str, prefix: str, max_uploads: int = 1
) -> ListMultipartUploadsOutputTypeDef:
    s3_client = session.client("s3", region_name=region)
    return s3_client.list_multipart_uploads(
        Bucket=bucket, MaxUploads=max_uploads, Prefix=prefix
    )



import os
import re
import shutil
import subprocess
import sys
import unittest

import pytest

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from toil.test import ToilTest, needs_docker
from toil.version import python


class ToilDocumentationTest(ToilTest):
    """Tests for scripts in the toil tutorials."""

    @classmethod
    def setUpClass(cls):
        super(ToilTest, cls).setUpClass()
        cls.directory = os.path.dirname(os.path.abspath(__file__))

    def tearDown(self) -> None:
        super(ToilTest, self).tearDown()

        jobstores = ["/mnt/ephemeral/workspace/toil-pull-requests/toilWorkflowRun"]
        for jobstore in jobstores:
            if os.path.exists(jobstore):
                shutil.rmtree(jobstore)

        unittest.TestCase.tearDown(self)

    """Just check the exit code"""

    def checkExitCode(self, script, extra_args: list[str] = []):
        program = os.path.join(self.directory, "scripts", script)
        process = subprocess.Popen(
            [python, program, "file:my-jobstore", "--clean=always"] + extra_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        if isinstance(stdout, bytes):
            stdout = stdout.decode("utf-8")
            stderr = stderr.decode("utf-8")
        if not process.returncode == 0:
            raise RuntimeError(stderr)
        return stdout + " " + stderr

    """Check the exit code and the output"""

    def checkExpectedOut(self, script, expectedOutput):
        outerr = self.checkExitCode(script)

        # Check that the expected output is there
        index = outerr.find(expectedOutput)
        self.assertGreater(index, -1, f"Expected:\n{expectedOutput}\nOutput:\n{outerr}")

    """Check the exit code and look for a pattern"""

    def checkExpectedPattern(self, script, expectedPattern):
        outerr = self.checkExitCode(script)

        # Check that the expected output pattern is there
        pattern = re.compile(expectedPattern, re.DOTALL)
        n = re.search(pattern, outerr)
        self.assertNotEqual(n, None, f"Pattern:\n{expectedPattern}\nOutput:\n{outerr}")

    def testStats(self):
        # This script asks for 4 cores but we might need to run the tests in only 3.
        self.checkExitCode("tutorial_stats.py", ["--scale=0.5"])

    @pytest.mark.timeout(1200)
    def testDynamic(self):
        self.checkExitCode("tutorial_dynamic.py")

    def testEncapsulation(self):
        self.checkExitCode("tutorial_encapsulation.py")

    def testEncapsulation2(self):
        self.checkExitCode("tutorial_encapsulation2.py")

    def testHelloworld(self):
        self.checkExpectedOut(
            "tutorial_helloworld.py", "Hello, world!, here's a message: You did it!\n"
        )

    def testInvokeworkflow(self):
        self.checkExpectedOut(
            "tutorial_invokeworkflow.py", "Hello, world!, here's a message: Woot\n"
        )

    def testInvokeworkflow2(self):
        self.checkExpectedOut(
            "tutorial_invokeworkflow2.py", "Hello, world!, I have a message: Woot!\n"
        )

    def testJobFunctions(self):
        self.checkExpectedOut(
            "tutorial_jobfunctions.py", "Hello world, I have a message: Woot!\n"
        )

    def testManaging(self):
        self.checkExitCode("tutorial_managing.py")

    def testManaging2(self):
        self.checkExitCode("tutorial_managing2.py")

    def testMultiplejobs(self):
        self.checkExpectedPattern(
            "tutorial_multiplejobs.py",
            "Hello world, I have a message: first.*Hello world, I have a message: "
            "second or third.*Hello world, I have a message: second or third.*Hello world,"
            " I have a message: last",
        )

    def testMultiplejobs2(self):
        self.checkExpectedPattern(
            "tutorial_multiplejobs2.py",
            "Hello world, I have a message: first.*Hello world, I have a message: "
            "second or third.*Hello world, I have a message: second or third.*Hello world,"
            " I have a message: last",
        )

    def testMultiplejobs3(self):
        self.checkExpectedPattern(
            "tutorial_multiplejobs3.py",
            "Hello world, I have a message: first.*Hello world, I have a message: "
            "second or third.*Hello world, I have a message: second or third.*Hello world,"
            " I have a message: last",
        )

    @pytest.mark.timeout(1200)
    def testPromises2(self):
        self.checkExpectedOut(
            "tutorial_promises2.py",
            "['00000', '00001', '00010', '00011', '00100', '00101', '00110', '00111',"
            " '01000', '01001', '01010', '01011', '01100', '01101', '01110', '01111',"
            " '10000', '10001', '10010', '10011', '10100', '10101', '10110', '10111',"
            " '11000', '11001', '11010', '11011', '11100', '11101', '11110', '11111']",
        )

    def testQuickstart(self):
        self.checkExpectedOut(
            "tutorial_quickstart.py", "Hello, world!, here's a message: Woot\n"
        )

    def testRequirements(self):
        self.checkExitCode("tutorial_requirements.py")

    def testArguments(self):
        self.checkExpectedOut(
            "tutorial_arguments.py", "Hello, world!, here's a message: Woot"
        )

    @needs_docker
    def testDocker(self):
        self.checkExitCode("tutorial_docker.py")

    def testPromises(self):
        self.checkExpectedPattern("tutorial_promises.py", "i is: 1.*i is: 2.*i is: 3")

    def testServices(self):
        self.checkExitCode("tutorial_services.py")

    def testStaging(self):
        self.checkExitCode("tutorial_staging.py")


if __name__ == "__main__":
    unittest.main()



import logging
import os
import subprocess
import sys

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp
from toil.test import get_data
from toil.version import python

logger = logging.getLogger(__name__)

"""
This workflow's purpose is to create files and jobs for viewing using stats,
status, and printDot() in toilDebugTest.py.  It's intended for future use in a
debugging tutorial containing a broken job.  It is also a minor integration test.
"""


def initialize_jobs(job):
    """
    Stub function used to start a toil workflow since toil workflows can only
    start with one job (but afterwards can run many in parallel).
    """
    job.fileStore.log_to_leader("""initialize_jobs""")


def writeA(job, mkFile):
    """Runs a program, and writes a string 'A' into A.txt using mkFile.py."""
    job.fileStore.log_to_leader("""writeA""")

    # temp folder for the run
    tempDir = job.fileStore.getLocalTempDir()

    # import files
    mkFile_fs = job.fileStore.readGlobalFile(
        mkFile[0], userPath=os.path.join(tempDir, mkFile[1])
    )

    # make a file (A.txt) and writes a string 'A' into it using 'mkFile.py'
    content = "A"
    cmd = python + " " + mkFile_fs + " " + "A.txt" + " " + content
    this_process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    this_process.wait()

    # get the output file and return it as a tuple of location + name
    output_filename = "A.txt"
    output_file = job.fileStore.writeGlobalFile(output_filename)
    A1 = (output_file, output_filename)
    rvDict = {"A1": A1}
    return rvDict


def writeB(job, mkFile, B_file):
    """
    Runs a program, extracts a string 'B' from an existing file, B_file.txt, and
    writes it into B.txt using mkFile.py.
    """
    job.fileStore.log_to_leader("""writeB""")

    # temp folder for the run
    tempDir = job.fileStore.getLocalTempDir()

    # import files
    mkFile_fs = job.fileStore.readGlobalFile(
        mkFile[0], userPath=os.path.join(tempDir, mkFile[1])
    )
    B_file_fs = job.fileStore.readGlobalFile(
        B_file[0], userPath=os.path.join(tempDir, B_file[1])
    )

    # make a file (B.txt) and write the contents of 'B_file.txt' into it using 'mkFile.py'
    with open(B_file_fs) as f:
        file_contents = ""
        for line in f:
            file_contents = file_contents + line

    cmd = python + " " + mkFile_fs + " " + "B.txt" + " " + file_contents
    this_process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    this_process.wait()

    # get the output file and return it as a tuple of location + name
    output_filename = "B.txt"
    output_file = job.fileStore.writeGlobalFile(output_filename)
    B1 = (output_file, output_filename)
    rvDict = {"B1": B1}
    return rvDict


def writeC(job):
    """Creates/writes a file, C.txt, containing the string 'C'."""
    job.fileStore.log_to_leader("""writeC""")

    # temp folder for the run
    tempDir = job.fileStore.getLocalTempDir()

    # get the output file and return it as a tuple of location + name
    output_filename = os.path.join(tempDir, "C.txt")
    with open(output_filename, "w") as f:
        f.write("C")
    output_file = job.fileStore.writeGlobalFile(output_filename)
    C1 = (output_file, "C.txt")
    rvDict = {"C1": C1}
    return rvDict


def writeABC(job, A_dict, B_dict, C_dict, filepath: str) -> None:
    """Takes 3 files (specified as dictionaries) and writes their contents to ABC.txt."""
    job.fileStore.log_to_leader("""writeABC""")

    # temp folder for the run
    tempDir = job.fileStore.getLocalTempDir()

    # import files
    A_fs = job.fileStore.readGlobalFile(
        A_dict["A1"][0], userPath=os.path.join(tempDir, A_dict["A1"][1])
    )
    B_fs = job.fileStore.readGlobalFile(
        B_dict["B1"][0], userPath=os.path.join(tempDir, B_dict["B1"][1])
    )
    C_fs = job.fileStore.readGlobalFile(
        C_dict["C1"][0], userPath=os.path.join(tempDir, C_dict["C1"][1])
    )

    file_contents = ""
    with open(A_fs) as f:
        for line in f:
            file_contents = file_contents + line

    with open(B_fs) as f:
        for line in f:
            file_contents = file_contents + line

    with open(C_fs) as f:
        for line in f:
            file_contents = file_contents + line

    with open(os.path.join(tempDir, "ABC.txt"), "w") as f:
        f.write(file_contents)

    # get the output file and return it as a tuple of location + name
    output_filename = os.path.join(tempDir, "ABC.txt")
    output_file = job.fileStore.writeGlobalFile(output_filename)
    job.fileStore.export_file(output_file, "file://" + filepath)


def finalize_jobs(job, num):
    """Does nothing but should be recorded in stats, status, and printDot()."""
    job.fileStore.log_to_leader("""finalize_jobs""")


def broken_job(job, num):
    """A job that will always fail.  To be used for a tutorial."""
    job.fileStore.log_to_leader("""broken_job""")
    file = toil.importFile(None)


if __name__ == "__main__":
    jobStorePath = sys.argv[1] if len(sys.argv) > 1 else mkdtemp("debugWorkflow")
    options = Job.Runner.getDefaultOptions(jobStorePath)
    options.clean = "never"
    options.stats = True
    options.logLevel = "INFO"
    with Toil(options) as toil:

        with get_data("test/utils/ABCWorkflowDebug/B_file.txt") as B_file_path:
            B_file0 = toil.importFile(B_file_path.as_uri())
        B_file0_preserveThisFilename = "B_file.txt"
        B_file = (B_file0, B_file0_preserveThisFilename)

        with get_data("test/utils/ABCWorkflowDebug/mkFile.py") as mkFile_path:
            file_maker0 = toil.importFile(mkFile_path.as_uri())
        file_maker0_preserveThisFilename = "mkFile.py"
        file_maker = (file_maker0, file_maker0_preserveThisFilename)

        job0 = Job.wrapJobFn(initialize_jobs)
        job1 = Job.wrapJobFn(writeA, file_maker)
        job2 = Job.wrapJobFn(writeB, file_maker, B_file)
        job3 = Job.wrapJobFn(writeC)
        with get_data("test/utils/ABCWorkflowDebug/ABC.txt") as filepath:
            job4 = Job.wrapJobFn(
                writeABC, job1.rv(), job2.rv(), job3.rv(), str(filepath)
            )
        job5 = Job.wrapJobFn(finalize_jobs, 1)
        job6 = Job.wrapJobFn(finalize_jobs, 2)
        job7 = Job.wrapJobFn(finalize_jobs, 3)
        job8 = Job.wrapJobFn(finalize_jobs, 4)

        # write files 'A.txt', 'B.txt', and 'C.txt'
        job0.addChild(job1)
        job1.addChild(job2)
        job2.addChild(job3)

        # finally use 'A.txt', 'B.txt', and 'C.txt' to write ABC.txt
        job0.addFollowOn(job4)

        # these jobs do nothing, but should display in status
        job4.addChild(job5)
        job4.addChild(job6)
        job4.addChild(job7)
        job4.addChild(job8)

        toil.start(job0)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fnmatch
import math
import os
import resource
import sys

from toil.lib.misc import StrPath

class ResourceMonitor:
    """
    Global resource monitoring widget.

    Presents class methods to get the resource usage of this process and child
    processes, and other class methods to adjust the statistics so they can
    account for e.g. resources used inside containers, or other resource usage
    that *should* be billable to the current process.
    """

    # Store some extra usage to tack onto the stats as module-level globals
    _extra_cpu_seconds: float = 0
    _extra_memory_ki: int = 0

    @classmethod
    def record_extra_memory(cls, peak_ki: int) -> None:
        """
        Become responsible for the given peak memory usage, in kibibytes.

        The memory will be treated as if it was used by a child process at the time
        our real child processes were also using their peak memory.
        """
        cls._extra_memory_ki = max(cls._extra_memory_ki, peak_ki)

    @classmethod
    def record_extra_cpu(cls, seconds: float) -> None:
        """
        Become responsible for the given CPU time.

        The CPU time will be treated as if it had been used by a child process.
        """
        cls._extra_cpu_seconds += seconds

    @classmethod
    def get_total_cpu_time_and_memory_usage(cls) -> tuple[float, int]:
        """
        Gives the total cpu time of itself and all its children, and the maximum RSS memory usage of
        itself and its single largest child (in kibibytes).
        """
        me = resource.getrusage(resource.RUSAGE_SELF)
        children = resource.getrusage(resource.RUSAGE_CHILDREN)
        total_cpu_time = (
            me.ru_utime
            + me.ru_stime
            + children.ru_utime
            + children.ru_stime
            + cls._extra_cpu_seconds
        )
        total_memory_usage = me.ru_maxrss + children.ru_maxrss
        if sys.platform == "darwin":
            # On Linux, getrusage works in "kilobytes" (really kibibytes), but on
            # Mac it works in bytes. See
            # <https://github.com/python/cpython/issues/74698>
            total_memory_usage = int(math.ceil(total_memory_usage / 1024))
        total_memory_usage += cls._extra_memory_ki
        return total_cpu_time, total_memory_usage

    @classmethod
    def get_total_cpu_time(cls) -> float:
        """Gives the total cpu time, including the children."""
        me = resource.getrusage(resource.RUSAGE_SELF)
        childs = resource.getrusage(resource.RUSAGE_CHILDREN)
        return (
            me.ru_utime
            + me.ru_stime
            + childs.ru_utime
            + childs.ru_stime
            + cls._extra_cpu_seconds
        )


def glob(glob_pattern: str, directoryname: StrPath) -> list[str]:
    """
    Walks through a directory and its subdirectories looking for files matching
    the glob_pattern and returns a list=[].

    :param directoryname: Any accessible folder name on the filesystem.
    :param glob_pattern: A string like ``*.txt``, which would find all text files.
    :return: A list=[] of absolute filepaths matching the glob pattern.
    """
    matches = []
    for root, dirnames, filenames in os.walk(directoryname):
        for filename in fnmatch.filter(filenames, glob_pattern):
            absolute_filepath = os.path.join(root, filename)
            matches.append(absolute_filepath)
    return matches



"""Neutral place for exceptions, to break import cycles."""

import logging
from typing import TYPE_CHECKING

from toil.statsAndLogging import StatsAndLogging

if TYPE_CHECKING:
    from toil.job import JobDescription
    from toil.jobStores.abstractJobStore import AbstractJobStore

logger = logging.getLogger(__name__)


class FailedJobsException(Exception):
    def __init__(
        self,
        job_store: "AbstractJobStore",
        failed_jobs: list["JobDescription"],
        exit_code: int = 1,
    ):
        """
        Make an exception to report failed jobs.

        :param job_store: The job store with the failed jobs in it.
        :param failed_jobs: All the failed jobs.
        :param exit_code: Recommended process exit code.
        """
        self.msg = (
            f"The job store '{job_store.locator}' contains "
            f"{len(failed_jobs)} failed jobs"
        )
        self.exit_code = exit_code
        try:
            self.msg += ": %s" % ", ".join(str(failedJob) for failedJob in failed_jobs)
            for job_desc in failed_jobs:
                if job_desc.logJobStoreFileID:
                    with job_desc.getLogFileHandle(job_store) as f:
                        self.msg += "\n" + StatsAndLogging.formatLogStream(
                            f, f'Log from job "{job_desc}"'
                        )
        # catch failures to prepare more complex details and only return the basics
        except Exception:
            logger.exception("Exception when compiling information about failed jobs")
        self.msg = self.msg.rstrip("\n")
        super().__init__()

        # Save fields that catchers can look at
        self.jobStoreLocator = job_store.locator
        self.numberOfFailedJobs = len(failed_jobs)

    def __str__(self) -> str:
        """Stringify the exception, including the message."""
        return self.msg



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from uuid import uuid4

import boto3
from moto import mock_aws

from toil.lib.aws import iam
from toil.test import ToilTest, needs_aws_s3

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class IAMTest(ToilTest):
    """Check that given permissions and associated functions perform correctly"""

    def test_permissions_iam(self):
        granted_perms = {
            "*": {"Action": ["ec2:*", "iam:*", "s3:*", "sdb:*"], "NotAction": []}
        }
        assert (
            iam.policy_permissions_allow(
                granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS
            )
            is True
        )
        granted_perms = {"*": {"Action": [], "NotAction": ["s3:*"]}}
        assert (
            iam.policy_permissions_allow(
                granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS
            )
            is True
        )

    def test_negative_permissions_iam(self):
        granted_perms = {"*": {"Action": ["ec2:*", "s3:*", "sdb:*"], "NotAction": []}}
        assert (
            iam.policy_permissions_allow(
                granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS
            )
            is False
        )
        granted_perms = {"*": {"Action": [], "NotAction": ["iam:*", "ec2:*"]}}
        assert (
            iam.policy_permissions_allow(
                granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS
            )
            is False
        )

    def test_wildcard_handling(self):
        assert iam.permission_matches_any("iam:CreateRole", ["iam:Create**"]) is True
        assert iam.permission_matches_any("iam:GetUser", ["iam:???????"]) is True
        assert iam.permission_matches_any("iam:ListRoleTags", ["iam:*?*Tags"]) is True
        assert iam.permission_matches_any("iam:*", ["*"]) is True
        assert iam.permission_matches_any("ec2:*", ["iam:*"]) is False

    @mock_aws
    @needs_aws_s3  # mock is incomplete, this avoid 'botocore.exceptions.NoCredentialsError: Unable to locate credentials'
    def test_get_policy_permissions(self):
        mock_iam = boto3.client("iam")

        # username that moto pretends we have from client.get_user()
        user_name = "default_user"
        mock_iam.create_user(UserName=user_name)

        group_name = "default_group"
        mock_iam.create_group(GroupName=group_name)

        mock_iam.add_user_to_group(GroupName=group_name, UserName=user_name)

        policy_response = mock_iam.create_policy(
            PolicyName="test_iam_createrole",
            PolicyDocument=json.dumps(
                {
                    "Version": "2012-10-17",  # represents version language
                    "Statement": [
                        {"Effect": "Allow", "Action": "iam:CreateRole", "Resource": "*"}
                    ],
                }
            ),
        )
        # attached user policy
        mock_iam.attach_user_policy(
            UserName=user_name, PolicyArn=policy_response["Policy"]["Arn"]
        )

        # inline user policy
        mock_iam.put_user_policy(
            UserName=user_name,
            PolicyName="test_iam_createinstanceprofile",
            PolicyDocument=json.dumps(
                {
                    "Version": "2012-10-17",  # represents version language
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": "iam:CreateInstanceProfile",
                            "Resource": "*",
                        }
                    ],
                }
            ),
        )

        # group policies
        policy_response = mock_iam.create_policy(
            PolicyName="test_iam_taginstanceprofile",
            PolicyDocument=json.dumps(
                {
                    "Version": "2012-10-17",  # represents version language
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": "iam:TagInstanceProfile",
                            "Resource": "*",
                        }
                    ],
                }
            ),
        )
        # attached group policy
        mock_iam.attach_group_policy(
            GroupName=group_name, PolicyArn=policy_response["Policy"]["Arn"]
        )

        # inline group policy
        mock_iam.put_group_policy(
            GroupName=group_name,
            PolicyName="test_iam_deleterole",
            PolicyDocument=json.dumps(
                {
                    "Version": "2012-10-17",  # represents version language
                    "Statement": [
                        {"Effect": "Allow", "Action": "iam:DeleteRole", "Resource": "*"}
                    ],
                }
            ),
        )

        actions_collection = iam.get_policy_permissions("us-west-2")

        actions_set = set(actions_collection["*"]["Action"])
        notactions_set = set(actions_collection["*"]["NotAction"])

        expected_actions = {
            "iam:CreateRole",
            "iam:CreateInstanceProfile",
            "iam:TagInstanceProfile",
            "iam:DeleteRole",
        }
        assert actions_set == expected_actions
        assert notactions_set == set()

    @needs_aws_s3
    def test_create_delete_iam_role(self):
        region = "us-west-2"
        role_name = f'test{str(uuid4()).replace("-", "")}'
        with self.subTest("Create role w/policies."):
            ec2_role_policy_document = json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Principal": {"Service": ["ec2.amazonaws.com"]},
                            "Action": ["sts:AssumeRole"],
                        }
                    ],
                }
            )
            policy = dict(
                s3_deny=dict(
                    Version="2012-10-17",
                    Statement=[dict(Effect="Deny", Resource="*", Action="s3:*")],
                )
            )
            iam.create_iam_role(
                role_name=role_name,
                assume_role_policy_document=ec2_role_policy_document,
                policies=policy,
                region=region,
            )
        with self.subTest("Delete role w/policies."):
            iam.delete_iam_role(role_name=role_name, region=region)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
from pathlib import Path
import signal
from typing import Optional

from toil.common import Toil
from toil.exceptions import FailedJobsException
from toil.job import Job
from toil.test import pslow as slow

import pytest

logger = logging.getLogger(__name__)


class TestRestartDAG:
    """
    Tests that restarted job DAGs don't run children of jobs that failed in the first run till the
    parent completes successfully in the restart.
    """

    @slow
    @pytest.mark.slow
    def testRestartedWorkflowSchedulesCorrectJobsOnFailedParent(
        self, tmp_path: Path
    ) -> None:
        self._testRestartedWorkflowSchedulesCorrectJobs(tmp_path, "raise")

    @slow
    @pytest.mark.slow
    def testRestartedWorkflowSchedulesCorrectJobsOnKilledParent(
        self, tmp_path: Path
    ) -> None:
        self._testRestartedWorkflowSchedulesCorrectJobs(tmp_path, "kill")

    def _testRestartedWorkflowSchedulesCorrectJobs(
        self, tmp_path: Path, failType: str
    ) -> None:
        """
        Creates a diamond DAG
            /->passingParent-\
        root                 |-->child
           \\->failingParent--/

        where root and passingParent are guaranteed to pass, while failingParent will fail.
        child should not run on start or restart and we assert that by ensuring that a file create
        iff child is run is not present on the system.

        :param str failType: Does failingParent fail on an assertionError, or is it killed.
        """
        # Specify options
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.logLevel = "DEBUG"
        options.retryCount = 0
        options.clean = "never"

        parentFile = tmp_path / "parent"
        childFile = tmp_path / "child"

        # Make the first job
        root = Job.wrapJobFn(passingFn)
        passingParent = Job.wrapJobFn(passingFn)
        failingParent = Job.wrapJobFn(failingFn, failType=failType, file=parentFile)
        child = Job.wrapJobFn(passingFn, file=childFile)

        # define the DAG
        root.addChild(passingParent)
        root.addChild(failingParent)
        passingParent.addChild(child)
        failingParent.addChild(child)

        failReasons = []

        assert not childFile.exists()

        errorRaised: Optional[BaseException] = None
        # Run the test
        for runMode in "start", "restart":
            errorRaised = None
            try:
                with Toil(options) as toil:
                    if runMode == "start":
                        toil.start(root)
                    else:
                        toil.restart()
            except Exception as e:
                logger.exception(e)
                errorRaised = e
            finally:
                # The processing of an AssertionError and FailedJobsException is the same so we do
                # it together in this finally clause.
                if errorRaised is not None:
                    if not parentFile.exists():
                        failReasons.append(
                            'The failing parent file did not exist on toil "%s".'
                            % runMode
                        )
                    if childFile.exists():
                        failReasons.append(
                            "The child file existed.  i.e. the child was run on "
                            'toil "%s".' % runMode
                        )
                    if isinstance(errorRaised, FailedJobsException):
                        if errorRaised.numberOfFailedJobs != 3:
                            failReasons.append(
                                'FailedJobsException was raised on toil "%s" but '
                                "the number of failed jobs (%s) was not 3."
                                % (runMode, errorRaised.numberOfFailedJobs)
                            )
                    elif isinstance(errorRaised, AssertionError):
                        failReasons.append(
                            "Toil raised an AssertionError instead of a "
                            'FailedJobsException on toil "%s".' % runMode
                        )
                    else:
                        failReasons.append("Toil raised error: %s" % errorRaised)
                    errorRaised = None
                    options.restart = True
                else:
                    pytest.fail('No errors were raised on toil "%s".' % runMode)
        if failReasons:
            pytest.fail(
                "Test failed for ({}) reasons:\n\t{}".format(
                    len(failReasons), "\n\t".join(failReasons)
                )
            )


def passingFn(job: Job, file: Optional[Path] = None) -> None:
    """
    This function is guaranteed to pass as it does nothing out of the ordinary.

    If "file" is provided, it will be created.

    :param file: The path of a file that must be created if provided.
    """
    if file is not None:
        # Emulates system touch.
        file.open("w").close()


def failingFn(job: Job, failType: str, file: Path) -> None:
    """
    This function is guaranteed to fail via a raised assertion, or an os.kill

    :param job: Job
    :param failType: 'raise' or 'kill
    :param file: The path of a file that must be created.
    """
    assert failType in ("raise", "kill")
    # Use that function to avoid code redundancy
    passingFn(job, file)

    if failType == "raise":
        assert False
    else:
        os.kill(os.getpid(), signal.SIGKILL)



import unittest
from uuid import uuid4

import pytest

from toil.provisioners import cluster_factory
from toil.test import integrative, slow
from toil.test.provisioners.clusterTest import AbstractClusterTest
from toil.test.wdl.wdltoil_test import (
    WDL_CONFORMANCE_TEST_COMMIT,
    WDL_CONFORMANCE_TEST_REPO,
)


@integrative
@slow
@pytest.mark.timeout(1800)
class WDLKubernetesClusterTest(AbstractClusterTest):
    """
    Ensure WDL works on the Kubernetes batchsystem.
    """

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.clusterName = "wdl-integration-test-" + str(uuid4())
        # t2.medium is the minimum t2 instance that permits Kubernetes
        self.leaderNodeType = "t2.medium"
        self.instanceTypes = ["t2.medium"]
        self.clusterType = "kubernetes"

    def setUp(self) -> None:
        super().setUp()
        self.jobStore = f"aws:{self.awsRegion()}:wdl-test-{uuid4()}"

    def launchCluster(self) -> None:
        self.createClusterUtil(
            args=[
                "--leaderStorage",
                str(self.requestedLeaderStorage),
                "--nodeTypes",
                ",".join(self.instanceTypes),
                "-w",
                ",".join(self.numWorkers),
                "--nodeStorage",
                str(self.requestedLeaderStorage),
            ]
        )

    def test_wdl_kubernetes_cluster(self) -> None:
        """
        Test that a wdl workflow works on a kubernetes cluster. Launches a cluster with 1 worker. This runs a wdl
        workflow that performs an image pull on the worker.
        :return:
        """
        self.numWorkers = "1"
        self.requestedLeaderStorage = 30
        # create the cluster
        self.launchCluster()
        # get leader
        self.cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )
        self.leader = self.cluster.getLeader()

        wdl_dir = "wdl_conformance_tests"

        # get the wdl-conformance-tests repo to get WDL tasks to run
        self.sshUtil(
            [
                "bash",
                "-c",
                f"git clone {WDL_CONFORMANCE_TEST_REPO} {wdl_dir} && cd {wdl_dir} && git checkout {WDL_CONFORMANCE_TEST_COMMIT}",
            ]
        )

        # run on kubernetes batchsystem
        toil_options = ["--batchSystem=kubernetes", f"--jobstore={self.jobStore}"]

        # run WDL workflow that will run singularity
        test_options = [f"tests/md5sum/md5sum.wdl", f"tests/md5sum/md5sum.json"]
        self.sshUtil(
            [
                "bash",
                "-c",
                f"cd {wdl_dir} && toil-wdl-runner {' '.join(test_options)} {' '.join(toil_options)}",
            ]
        )


if __name__ == "__main__":
    unittest.main()  # run all tests



from __future__ import absolute_import
from six.moves import xrange
from configargparse import ArgumentParser
import os
import logging
import random

from toil.common import Toil
from toil.job import Job


def setup(job, input_file_id, n, down_checkpoints):
    """Sets up the sort.
    Returns the FileID of the sorted file
    """
    # Write the input file to the file store
    job.fileStore.log_to_leader("Starting the merge sort")
    return job.addChildJobFn(down,
                             input_file_id, n,
                             down_checkpoints=down_checkpoints,
                             memory='600M').rv()


def down(job, input_file_id, n, down_checkpoints):
    """Input is a file and a range into that file to sort and an output location in which
    to write the sorted file.
    If the range is larger than a threshold N the range is divided recursively and
    a follow on job is then created which merges back the results. Otherwise,
    the file is sorted and placed in the output.
    """
    # Read the file
    input_file = job.fileStore.readGlobalFile(input_file_id, cache=False)
    length = os.path.getsize(input_file)
    if length > n:
        # We will subdivide the file
        job.fileStore.log_to_leader("Splitting file: %s of size: %s"
                                  % (input_file_id, length), level=logging.CRITICAL)
        # Split the file into two copies
        mid_point = get_midpoint(input_file, 0, length)
        t1 = job.fileStore.getLocalTempFile()
        with open(t1, 'w') as fH:
            copy_subrange_of_file(input_file, 0, mid_point + 1, fH)
        t2 = job.fileStore.getLocalTempFile()
        with open(t2, 'w') as fH:
            copy_subrange_of_file(input_file, mid_point + 1, length, fH)

        # Call the down function recursively
        return job.addFollowOnJobFn(up, job.addChildJobFn(down, job.fileStore.writeGlobalFile(t1), n,
                                    down_checkpoints=down_checkpoints, memory='600M').rv(),
                                    job.addChildJobFn(down, job.fileStore.writeGlobalFile(t2), n,
                                                      down_checkpoints=down_checkpoints,
                                                      memory='600M').rv()).rv()
    else:
        # We can sort this bit of the file
        job.fileStore.log_to_leader("Sorting file: %s of size: %s"
                                  % (input_file_id, length), level=logging.CRITICAL)
        # Sort the copy and write back to the fileStore
        output_file = job.fileStore.getLocalTempFile()
        sort(input_file, output_file)
        return job.fileStore.writeGlobalFile(output_file)


def up(job, input_file_id_1, input_file_id_2):
    """Merges the two files and places them in the output.
    """
    with job.fileStore.writeGlobalFileStream() as (fileHandle, output_id):
        with job.fileStore.readGlobalFileStream(input_file_id_1) as inputFileHandle1:
            with job.fileStore.readGlobalFileStream(input_file_id_2) as inputFileHandle2:
                job.fileStore.log_to_leader("Merging %s and %s to %s"
                                          % (input_file_id_1, input_file_id_2, output_id))
                merge(inputFileHandle1, inputFileHandle2, fileHandle)

        # Cleanup up the input files - these deletes will occur after the completion is successful.
        job.fileStore.deleteGlobalFile(input_file_id_1)
        job.fileStore.deleteGlobalFile(input_file_id_2)
        return output_id


# convenience functions
def sort(in_file, out_file):
    """Sorts the given file.
    """
    filehandle = open(in_file, 'r')
    lines = filehandle.readlines()
    filehandle.close()
    lines.sort()
    filehandle = open(out_file, 'w')
    for line in lines:
        filehandle.write(line)
    filehandle.close()


def merge(filehandle_1, filehandle_2, output_filehandle):
    """Merges together two files maintaining sorted order.
    """
    line2 = filehandle_2.readline()
    for line1 in filehandle_1.readlines():
        while line2 != '' and line2 <= line1:
            output_filehandle.write(line2)
            line2 = filehandle_2.readline()
        output_filehandle.write(line1)
    while line2 != '':
        output_filehandle.write(line2)
        line2 = filehandle_2.readline()


def copy_subrange_of_file(input_file, file_start, file_end, output_filehandle):
    """Copies the range (in bytes) between fileStart and fileEnd to the given
    output file handle.
    """
    with open(input_file, 'r') as fileHandle:
        fileHandle.seek(file_start)
        data = fileHandle.read(file_end - file_start)
        assert len(data) == file_end - file_start
        output_filehandle.write(data)


def get_midpoint(file, file_start, file_end):
    """Finds the point in the file to split.
    Returns an int i such that fileStart <= i < fileEnd
    """
    filehandle = open(file, 'r')
    mid_point = (file_start + file_end) / 2
    assert mid_point >= file_start
    filehandle.seek(mid_point)
    line = filehandle.readline()
    assert len(line) >= 1
    if len(line) + mid_point < file_end:
        return mid_point + len(line) - 1
    filehandle.seek(file_start)
    line = filehandle.readline()
    assert len(line) >= 1
    assert len(line) + file_start <= file_end
    return len(line) + file_start - 1


def make_file_to_sort(file_name, lines, line_length):
    with open(file_name, 'w') as fileHandle:
        for _ in xrange(lines):
            line = "".join(random.choice('actgACTGNXYZ') for _ in xrange(line_length - 1)) + '\n'
            fileHandle.write(line)


def main():
    parser = ArgumentParser()
    Job.Runner.addToilOptions(parser)

    parser.add_argument('--num-lines', default=1000, help='Number of lines in file to sort.', type=int)
    parser.add_argument('--line-length', default=50, help='Length of lines in file to sort.', type=int)
    parser.add_argument("--N",
                        help="The threshold below which a serial sort function is used to sort file. "
                        "All lines must of length less than or equal to N or program will fail",
                        default=10000)

    options = parser.parse_args()

    if int(options.N) <= 0:
        raise RuntimeError("Invalid value of N: %s" % options.N)

    file_name = 'file_to_sort.txt'
    make_file_to_sort(file_name=file_name, lines=options.num_lines, line_length=options.line_length)

    with Toil(options) as toil:
        sort_file_url = 'file://' + os.path.abspath('file_to_sort.txt')
        if not toil.options.restart:
            sort_file_id = toil.importFile(sort_file_url)
            sorted_file_id = toil.start(Job.wrapJobFn(setup, sort_file_id, int(options.N), False, memory='600M'))
        else:
            sorted_file_id = toil.restart()
        toil.exportFile(sorted_file_id, sort_file_url)

if __name__ == '__main__':
    main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Rsyncs into the toil appliance container running on the leader of the cluster."""
import argparse
import logging
import sys

from toil.common import parser_with_common_options
from toil.provisioners import NoSuchClusterException, cluster_factory
from toil.statsAndLogging import set_logging_from_options

logger = logging.getLogger(__name__)


def main() -> None:
    parser = parser_with_common_options(
        provisioner_options=True, jobstore_option=False, prog="toil rsync-cluster"
    )
    parser.add_argument(
        "--insecure",
        dest="insecure",
        action="store_true",
        required=False,
        help="Temporarily disable strict host key checking.",
    )
    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        help="Arguments to pass to"
        "`rsync`. Takes any arguments that rsync accepts. Specify the"
        " remote with a colon. For example, to upload `example.py`,"
        " specify `toil rsync-cluster -p aws test-cluster example.py :`."
        "\nOr, to download a file from the remote:, `toil rsync-cluster"
        " -p aws test-cluster :example.py .`",
    )
    options = parser.parse_args()
    set_logging_from_options(options)
    cluster = cluster_factory(
        provisioner=options.provisioner,
        clusterName=options.clusterName,
        zone=options.zone,
    )
    try:
        cluster.getLeader().coreRsync(args=options.args, strict=not options.insecure)
    except NoSuchClusterException as e:
        logger.error(e)
        sys.exit(1)



import os
import re
import sys
import textwrap
import types
from importlib.metadata import version as metadata_version
from typing import Any

from toil.version import version


def main() -> None:
    modules = loadModules()

    if len(sys.argv) < 2 or sys.argv[1] == "--help":
        printHelp(modules)
        sys.exit(0)

    cmd = sys.argv[1]
    if cmd == "--version":
        printVersion()
        sys.exit(0)

    try:
        module = modules[cmd]
    except KeyError:
        sys.stderr.write(
            f'Unknown option "{cmd}".  Pass --help to display usage information.\n'
        )
        sys.exit(1)

    del sys.argv[1]
    get_or_die(module, "main")()


def get_or_die(module: types.ModuleType, name: str) -> Any:
    """
    Get an object from a module or complain that it is missing.
    """
    if hasattr(module, name):
        return getattr(module, name)
    else:
        sys.stderr.write(
            f"Internal Toil error!\nToil utility module "
            f"{module.__name__} is missing required attribute {name}\n"
        )
        sys.exit(1)


def loadModules() -> dict[str, types.ModuleType]:
    # noinspection PyUnresolvedReferences
    from toil.utils import toilClean  # noqa
    from toil.utils import toilConfig  # noqa
    from toil.utils import toilDebugFile  # noqa
    from toil.utils import toilDebugJob  # noqa
    from toil.utils import toilDestroyCluster  # noqa
    from toil.utils import toilKill  # noqa
    from toil.utils import toilLaunchCluster  # noqa
    from toil.utils import toilRsyncCluster  # noqa
    from toil.utils import toilServer  # noqa
    from toil.utils import toilSshCluster  # noqa
    from toil.utils import toilStats  # noqa
    from toil.utils import toilStatus  # noqa

    return {
        "-".join([i.lower() for i in re.findall("[A-Z][^A-Z]*", name)]): module
        for name, module in locals().items()
    }


def printHelp(modules: dict[str, types.ModuleType]) -> None:
    name = os.path.basename(sys.argv[0])
    descriptions = "\n        ".join(
        f'{cmd} - {get_or_die(mod, "__doc__").strip()}'
        for cmd, mod in modules.items()
        if mod
    )
    print(
        textwrap.dedent(
            f"""
        Usage: {name} COMMAND ...
               {name} --help
               {name} COMMAND --help

        Where COMMAND is one of the following:

        {descriptions}
        """[
                1:
            ]
        )
    )


def printVersion() -> None:
    try:
        detected_version = metadata_version("toil")
        if "a" in detected_version:
            # This probably means Toil is installed as development
            print(version)
        else:
            print(detected_version)
    except:
        print(f"Version gathered from toil.version: {version}")



import datetime
import getpass
import logging
import os
import random
import socket
import subprocess
import sys
import time
from collections.abc import Iterator
from contextlib import closing
from typing import Optional, Union
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

logger = logging.getLogger(__name__)

StrPath: TypeAlias = Union[str, os.PathLike[str]]
FileDescriptorOrPath: TypeAlias = Union[int, bytes, os.PathLike[bytes], StrPath]

def get_public_ip() -> str:
    """Get the IP that this machine uses to contact the internet.

    If behind a NAT, this will still be this computer's IP, and not the router's."""
    try:
        # Try to get the internet-facing IP by attempting a connection
        # to a non-existent server and reading what IP was used.
        ip = "127.0.0.1"
        with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as sock:
            # 203.0.113.0/24 is reserved as TEST-NET-3 by RFC 5737, so
            # there is guaranteed to be no one listening on the other
            # end (and we won't accidentally DOS anyone).
            sock.connect(("203.0.113.1", 1))
            ip = sock.getsockname()[0]
        return ip
    except:
        # Something went terribly wrong. Just give loopback rather
        # than killing everything, because this is often called just
        # to provide a default argument
        return "127.0.0.1"


def get_user_name() -> str:
    """
    Get the current user name, or a suitable substitute string if the user name
    is not available.
    """
    try:
        try:
            return getpass.getuser()
        except KeyError:
            # This is expected if the user isn't in /etc/passwd, such as in a
            # Docker container when running as a weird UID. Make something up.
            return "UnknownUser" + str(os.getuid())
    except Exception as e:
        # We can't get the UID, or something weird has gone wrong.
        logger.error("Unexpected error getting user name: %s", e)
        return "UnknownUser"


def utc_now() -> datetime.datetime:
    """Return a datetime in the UTC timezone corresponding to right now."""
    return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)


def unix_now_ms() -> float:
    """Return the current time in milliseconds since the Unix epoch."""
    return time.time() * 1000

def unix_seconds_to_timestamp(timestamp: float) -> str:
    """
    Convert a time in seconds since the Unix epoch to an ISO 8601 string.
    """
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc).isoformat()

def unix_seconds_to_local_time(timestamp: float) -> datetime.datetime:
    """
    Returns a local time corresponding to the given number of seconds since the Unix epoch.
    """
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc).astimezone()

def seconds_to_duration(time_difference: float) -> str:
    """
    Convert a time difference in seconds to an ISO 8601 duration string.
    """
    return f"PT{time_difference:.3f}S"


def slow_down(seconds: float) -> float:
    """
    Toil jobs that have completed are not allowed to have taken 0 seconds, but
    Kubernetes timestamps round things to the nearest second. It is possible in
    some batch systems for a pod to have identical start and end timestamps.

    This function takes a possibly 0 job length in seconds and enforces a
    minimum length to satisfy Toil.

    :param float seconds: Timestamp difference

    :return: seconds, or a small positive number if seconds is 0
    :rtype: float
    """

    return max(seconds, sys.float_info.epsilon)


def printq(msg: str, quiet: bool, log: bool = False) -> None:
    """
    This is for functions used simultaneously in Toil proper and in the admin scripts.

    Our admin scripts "print" to stdout, while Toil proper uses logging.  For a script that,
    for example, cleans up IAM, EC2, etc. cruft leftover after failed CI runs, we can call
    an AWS delete IAM role function, and this prints or logs progress (unless quiet is True),
    depending on whether the function is called in, say, the jobstore or a script.

    :param msg: The string to print or log to stdout.
    :param quiet: Silent output to stdout.
    :param log: Use logging (else "print" to the screen).
    """
    if not quiet:
        if not log:
            print(msg)
        else:
            logger.debug(msg)


def truncExpBackoff() -> Iterator[float]:
    # as recommended here https://forums.aws.amazon.com/thread.jspa?messageID=406788#406788
    # and here https://cloud.google.com/storage/docs/xml-api/reference-status
    yield 0
    t = 1
    while t < 1024:
        # google suggests this dither
        yield t + random.random()
        t *= 2
    while True:
        yield t


class CalledProcessErrorStderr(subprocess.CalledProcessError):
    """Version of CalledProcessError that include stderr in the error message if it is set"""

    def __str__(self) -> str:
        if (self.returncode < 0) or (self.stderr is None):
            return str(super())
        else:
            err = (
                self.stderr
                if isinstance(self.stderr, str)
                else self.stderr.decode("ascii", errors="replace")
            )
            return "Command '%s' exit status %d: %s" % (self.cmd, self.returncode, err)


def call_command(
    cmd: list[str],
    *args: str,
    input: Optional[str] = None,
    timeout: Optional[float] = None,
    useCLocale: bool = True,
    env: Optional[dict[str, str]] = None,
    quiet: Optional[bool] = False
) -> str:
    """
    Simplified calling of external commands.

    If the process fails, CalledProcessErrorStderr is raised.

    The captured stderr is always printed, regardless of
    if an exception occurs, so it can be logged.

    Always logs the command at debug log level.

    :param quiet: If True, do not log the command output. If False (the
           default), do log the command output at debug log level.

    :param useCLocale: If True, C locale is forced, to prevent failures that
           can occur in some batch systems when using UTF-8 locale.

    :returns: Command standard output, decoded as utf-8.
    """

    # NOTE: Interface MUST be kept in sync with call_sacct and call_scontrol in
    # test_slurm.py, which monkey-patch this!

    # using non-C locales can cause GridEngine commands, maybe other to
    # generate errors
    if useCLocale:
        env = dict(os.environ) if env is None else dict(env)  # copy since modifying
        env["LANGUAGE"] = env["LC_ALL"] = "C"

    logger.debug("run command: {}".format(" ".join(cmd)))
    start_time = datetime.datetime.now()
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        errors="replace",
        env=env,
    )
    stdout, stderr = proc.communicate(input=input, timeout=timeout)
    end_time = datetime.datetime.now()
    runtime = (end_time - start_time).total_seconds()
    sys.stderr.write(stderr)
    if proc.returncode != 0:
        logger.debug(
            "command failed in {}s: {}: {}".format(
                runtime, " ".join(cmd), stderr.rstrip()
            )
        )
        raise CalledProcessErrorStderr(
            proc.returncode, cmd, output=stdout, stderr=stderr
        )
    logger.debug(
        "command succeeded in {}s: {}{}".format(
            runtime, " ".join(cmd), (": " + stdout.rstrip()) if not quiet else ""
        )
    )
    return stdout



from typing import Any, Callable, List, Mapping, Optional, Tuple, Union

from .listener import IListenerExcHandler, Listener, UserListener
from .notificationmgr import INotificationHandler
from .topicmgr import TopicManager, TreeConfig

TopicFilter = Callable[[str], bool]
ListenerFilter = Callable[[Listener], bool]

class Publisher:
    def __init__(self, treeConfig: TreeConfig = ...) -> None: ...
    def getTopicMgr(self) -> TopicManager: ...
    def getListenerExcHandler(self) -> IListenerExcHandler: ...
    def setListenerExcHandler(self, handler: IListenerExcHandler) -> None: ...
    def addNotificationHandler(self, handler: INotificationHandler) -> None: ...
    def clearNotificationHandlers(self) -> None: ...
    def setNotificationFlags(self, **kwargs: Mapping[str, Optional[bool]]) -> None: ...
    def getNotificationFlags(self) -> Mapping[str, bool]: ...
    def setTopicUnspecifiedFatal(
        self, newVal: bool = ..., checkExisting: bool = ...
    ) -> bool: ...
    def subscribe(
        self, listener: UserListener, topicName: str, **curriedArgs: Any
    ) -> Tuple[Listener, bool]: ...
    def unsubscribe(self, listener: UserListener, topicName: str) -> Listener: ...
    def unsubAll(
        self,
        topicName: str = ...,
        listenerFilter: ListenerFilter = ...,
        topicFilter: Union[str, TopicFilter] = ...,
    ) -> List[Listener]: ...
    def sendMessage(self, topicName: str, **msgData: Any) -> None: ...



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp


def globalFileStoreJobFn(job):
    job.log(
        "The following example exercises all the methods provided "
        "by the toil.fileStores.abstractFileStore.AbstractFileStore class"
    )

    # Create a local temporary file.
    scratchFile = job.fileStore.getLocalTempFile()

    # Write something in the scratch file.
    with open(scratchFile, "w") as fH:
        fH.write("What a tangled web we weave")

    # Write a copy of the file into the file-store; fileID is the key that can be used to retrieve the file.
    # This write is asynchronous by default
    fileID = job.fileStore.writeGlobalFile(scratchFile)

    # Write another file using a stream; fileID2 is the
    # key for this second file.
    with job.fileStore.writeGlobalFileStream(cleanup=True) as (fH, fileID2):
        fH.write(b"Out brief candle")

    # Now read the first file; scratchFile2 is a local copy of the file that is read-only by default.
    scratchFile2 = job.fileStore.readGlobalFile(fileID)

    # Read the second file to a desired location: scratchFile3.
    scratchFile3 = os.path.join(job.tempDir, "foo.txt")
    job.fileStore.readGlobalFile(fileID2, userPath=scratchFile3)

    # Read the second file again using a stream.
    with job.fileStore.readGlobalFileStream(fileID2) as fH:
        print(fH.read())  # This prints "Out brief candle"

    # Delete the first file from the global file-store.
    job.fileStore.deleteGlobalFile(fileID)

    # It is unnecessary to delete the file keyed by fileID2 because we used the cleanup flag,
    # which removes the file after this job and all its successors have run (if the file still exists)


if __name__ == "__main__":
    jobstore: str = mkdtemp("tutorial_managing2")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "INFO"
    options.clean = "always"

    with Toil(options) as toil:
        toil.start(Job.wrapJobFn(globalFileStoreJobFn))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import re
import socket
import sys
from datetime import datetime
from typing import TYPE_CHECKING, Optional

from toil.lib.web import web_session

from docker.errors import ImageNotFound
from toil.lib.memoize import memoize
from toil.lib.retry import retry as retry
from toil.version import currentCommit

if TYPE_CHECKING:
    from toil.common import Config

log = logging.getLogger(__name__)


def which(cmd, mode=os.F_OK | os.X_OK, path=None) -> Optional[str]:
    """
    Return the path with conforms to the given mode on the Path.

    [Copy-pasted in from python3.6's shutil.which().]

    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.

    :returns: The path found, or None.
    """

    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, mode):
        return os.path.exists(fn) and os.access(fn, mode) and not os.path.isdir(fn)

    # If we're given a path with a directory part, look it up directly rather
    # than referring to PATH directories. This includes checking relative to the
    # current directory, e.g. ./script
    if os.path.dirname(cmd):
        if _access_check(cmd, mode):
            return cmd
        return None

    if path is None:
        path = os.environ.get("PATH", os.defpath)
    if not path:
        return None
    path = path.split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if not os.curdir in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        # If it does match, only test that one, otherwise we have to try
        # others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir in path:
        normdir = os.path.normcase(dir)
        if not normdir in seen:
            seen.add(normdir)
            for thefile in files:
                name = os.path.join(dir, thefile)
                if _access_check(name, mode):
                    return name
    return None


def toilPackageDirPath() -> str:
    """
    Return the absolute path of the directory that corresponds to the top-level toil package.

    The return value is guaranteed to end in '/toil'.
    """
    result = os.path.dirname(os.path.realpath(__file__))
    if not result.endswith("/toil"):
        raise RuntimeError("The top-level toil package is not named Toil.")
    return result


def inVirtualEnv() -> bool:
    """Test if we are inside a virtualenv or Conda virtual environment."""
    return (
        "VIRTUAL_ENV" in os.environ
        or "CONDA_DEFAULT_ENV" in os.environ
        or hasattr(sys, "real_prefix")
        or (hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix)
    )


def resolveEntryPoint(entryPoint: str) -> str:
    """
    Find the path to the given entry point that *should* work on a worker.

    :returns: The path found, which may be an absolute or a relative path.
    """
    if os.environ.get("TOIL_CHECK_ENV", None) == "True" and inVirtualEnv():
        path = os.path.join(os.path.dirname(sys.executable), entryPoint)
        # Inside a virtualenv we try to use absolute paths to the entrypoints.
        if os.path.isfile(path):
            # If the entrypoint is present, Toil must have been installed into the virtualenv (as
            # opposed to being included via --system-site-packages). For clusters this means that
            # if Toil is installed in a virtualenv on the leader, it must be installed in
            # a virtualenv located at the same path on each worker as well.
            if not os.access(path, os.X_OK):
                raise RuntimeError(
                    "Cannot access the Toil virtualenv. If installed in a virtualenv on a cluster, make sure that the virtualenv path is the same for the leader and workers."
                )
            return path
    # Otherwise, we aren't in a virtualenv, or we're in a virtualenv but Toil
    # came in via --system-site-packages, or we think the virtualenv might not
    # exist on the workers.
    return entryPoint


@memoize
def physicalMemory() -> int:
    """
    Calculate the total amount of physical memory, in bytes.

    >>> n = physicalMemory()
    >>> n > 0
    True
    >>> n == physicalMemory()
    True
    """
    try:
        return os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
    except ValueError:
        import subprocess

        return int(
            subprocess.check_output(["sysctl", "-n", "hw.memsize"])
            .decode("utf-8")
            .strip()
        )


def physicalDisk(directory: str) -> int:
    diskStats = os.statvfs(directory)
    return diskStats.f_frsize * diskStats.f_bavail


def applianceSelf(forceDockerAppliance: bool = False) -> str:
    """
    Return the fully qualified name of the Docker image to start Toil appliance containers from.

    The result is determined by the current version of Toil and three environment variables:
    ``TOIL_DOCKER_REGISTRY``, ``TOIL_DOCKER_NAME`` and ``TOIL_APPLIANCE_SELF``.

    ``TOIL_DOCKER_REGISTRY`` specifies an account on a publicly hosted docker registry like Quay
    or Docker Hub. The default is UCSC's CGL account on Quay.io where the Toil team publishes the
    official appliance images. ``TOIL_DOCKER_NAME`` specifies the base name of the image. The
    default of `toil` will be adequate in most cases. ``TOIL_APPLIANCE_SELF`` fully qualifies the
    appliance image, complete with registry, image name and version tag, overriding both
    ``TOIL_DOCKER_NAME`` and `TOIL_DOCKER_REGISTRY`` as well as the version tag of the image.
    Setting TOIL_APPLIANCE_SELF will not be necessary in most cases.
    """
    import toil.version

    registry = lookupEnvVar(
        name="docker registry",
        envName="TOIL_DOCKER_REGISTRY",
        defaultValue=toil.version.dockerRegistry,
    )
    name = lookupEnvVar(
        name="docker name",
        envName="TOIL_DOCKER_NAME",
        defaultValue=toil.version.dockerName,
    )
    appliance = lookupEnvVar(
        name="docker appliance",
        envName="TOIL_APPLIANCE_SELF",
        defaultValue=registry + "/" + name + ":" + toil.version.dockerTag,
    )

    checkDockerSchema(appliance)

    if forceDockerAppliance:
        return appliance
    else:
        return checkDockerImageExists(appliance=appliance)


def customDockerInitCmd() -> str:
    """
    Return the custom command set by the ``TOIL_CUSTOM_DOCKER_INIT_COMMAND`` environment variable.

    The custom docker command is run prior to running the workers and/or the primary node's services.

    This can be useful for doing any custom initialization on instances (e.g. authenticating to
    private docker registries). Any single quotes are escaped and the command cannot contain a
    set of blacklisted chars (newline or tab).

    :returns: The custom command, or an empty string is returned if the environment variable is not set.
    """
    command = lookupEnvVar(
        name="user-defined custom docker init command",
        envName="TOIL_CUSTOM_DOCKER_INIT_COMMAND",
        defaultValue="",
    )
    _check_custom_bash_cmd(command)
    return command.replace("'", "'\\''")  # Ensure any single quotes are escaped.


def customInitCmd() -> str:
    """
    Return the custom command set by the ``TOIL_CUSTOM_INIT_COMMAND`` environment variable.

    The custom init command is run prior to running Toil appliance itself in workers and/or the
    primary node (i.e. this is run one stage before ``TOIL_CUSTOM_DOCKER_INIT_COMMAND``).

    This can be useful for doing any custom initialization on instances (e.g. authenticating to
    private docker registries). Any single quotes are escaped and the command cannot contain a
    set of blacklisted chars (newline or tab).

    returns: the custom command or n empty string is returned if the environment variable is not set.
    """
    command = lookupEnvVar(
        name="user-defined custom init command",
        envName="TOIL_CUSTOM_INIT_COMMAND",
        defaultValue="",
    )
    _check_custom_bash_cmd(command)
    return command.replace("'", "'\\''")  # Ensure any single quotes are escaped.


def _check_custom_bash_cmd(cmd_str):
    """Ensure that the Bash command doesn't contain invalid characters."""
    if re.search(r"[\n\r\t]", cmd_str):
        raise RuntimeError(
            f'"{cmd_str}" contains invalid characters (newline and/or tab).'
        )


def lookupEnvVar(name: str, envName: str, defaultValue: str) -> str:
    """
    Look up environment variables that control Toil and log the result.

    :param name: the human readable name of the variable
    :param envName: the name of the environment variable to lookup
    :param defaultValue: the fall-back value
    :return: the value of the environment variable or the default value the variable is not set
    """
    try:
        value = os.environ[envName]
    except KeyError:
        log.info(
            "Using default %s of %s as %s is not set.", name, defaultValue, envName
        )
        return defaultValue
    else:
        log.info(
            "Overriding %s of %s with %s from %s.", name, defaultValue, value, envName
        )
        return value


def checkDockerImageExists(appliance: str) -> str:
    """
    Attempt to check a url registryName for the existence of a docker image with a given tag.

    :param appliance: The url of a docker image's registry (with a tag) of the form:
                      'quay.io/<repo_path>:<tag>' or '<repo_path>:<tag>'.
                      Examples: 'quay.io/ucsc_cgl/toil:latest', 'ubuntu:latest', or
                      'broadinstitute/genomes-in-the-cloud:2.0.0'.
    :return: Raises an exception if the docker image cannot be found or is invalid.  Otherwise, it
             will return the appliance string.
    """
    if currentCommit in appliance:
        return appliance
    registryName, imageName, tag = parseDockerAppliance(appliance)

    if registryName == "docker.io":
        return requestCheckDockerIo(
            origAppliance=appliance, imageName=imageName, tag=tag
        )
    else:
        return requestCheckRegularDocker(
            origAppliance=appliance,
            registryName=registryName,
            imageName=imageName,
            tag=tag,
        )


def parseDockerAppliance(appliance: str) -> tuple[str, str, str]:
    """
    Derive parsed registry, image reference, and tag from a docker image string.

    Example: "quay.io/ucsc_cgl/toil:latest"
    Should return: "quay.io", "ucsc_cgl/toil", "latest"

    If a registry is not defined, the default is: "docker.io"
    If a tag is not defined, the default is: "latest"

    :param appliance: The full url of the docker image originally
                      specified by the user (or the default).
                      e.g. "quay.io/ucsc_cgl/toil:latest"
    :returns: registryName, imageName, tag
    """
    appliance = appliance.lower()

    # get the tag
    if ":" in appliance:
        tag = appliance.split(":")[-1]
        appliance = appliance[: -(len(":" + tag))]  # remove only the tag
    else:
        # default to 'latest' if no tag is specified
        tag = "latest"

    # get the registry and image
    registryName = "docker.io"  # default if not specified
    imageName = appliance  # will be true if not specified
    if "/" in appliance and "." in appliance.split("/")[0]:
        registryName = appliance.split("/")[0]
        imageName = appliance[len(registryName) :]
    registryName = registryName.strip("/")
    imageName = imageName.strip("/")

    return registryName, imageName, tag


def checkDockerSchema(appliance):
    if not appliance:
        raise ImageNotFound("No docker image specified.")
    elif "://" in appliance:
        raise ImageNotFound(
            "Docker images cannot contain a schema (such as '://'): %s" "" % appliance
        )
    elif len(appliance) > 256:
        raise ImageNotFound(
            "Docker image must be less than 256 chars: %s" "" % appliance
        )


class ApplianceImageNotFound(ImageNotFound):
    """
    Error raised when using TOIL_APPLIANCE_SELF results in an HTTP error.

    :param str origAppliance: The full url of the docker image originally
                              specified by the user (or the default).
                              e.g. "quay.io/ucsc_cgl/toil:latest"
    :param str url: The URL at which the image's manifest is supposed to appear
    :param int statusCode: the failing HTTP status code returned by the URL
    """

    def __init__(self, origAppliance, url, statusCode):
        msg = (
            "The docker image that TOIL_APPLIANCE_SELF specifies (%s) produced "
            "a nonfunctional manifest URL (%s). The HTTP status returned was %s. "
            "The specifier is most likely unsupported or malformed.  "
            "Please supply a docker image with the format: "
            "'<websitehost>.io/<repo_path>:<tag>' or '<repo_path>:<tag>' "
            "(for official docker.io images).  Examples: "
            "'quay.io/ucsc_cgl/toil:latest', 'ubuntu:latest', or "
            "'broadinstitute/genomes-in-the-cloud:2.0.0'."
            "" % (origAppliance, url, str(statusCode))
        )
        super().__init__(msg)


# Cache images we know exist so we don't have to ask the registry about them
# all the time.
KNOWN_EXTANT_IMAGES = set()


def requestCheckRegularDocker(
    origAppliance: str, registryName: str, imageName: str, tag: str
) -> bool:
    """
    Check if an image exists using the requests library.

    URL is based on the
    `docker v2 schema <https://docs.docker.com/registry/spec/manifest-v2-2/>`_.

    This has the following format: ``https://{websitehostname}.io/v2/{repo}/manifests/{tag}``

    Does not work with the official (docker.io) site, because they require an OAuth token, so a
    separate check is done for docker.io images.

    :param origAppliance: The full url of the docker image originally
        specified by the user (or the default). For example, ``quay.io/ucsc_cgl/toil:latest``.
    :param registryName: The url of a docker image's registry. For example, ``quay.io``.
    :param imageName: The image, including path and excluding the tag. For example, ``ucsc_cgl/toil``.
    :param tag: The tag used at that docker image's registry. For example, ``latest``.
    :raises: ApplianceImageNotFound if no match is found.
    :return: Return True if match found.
    """
    if origAppliance in KNOWN_EXTANT_IMAGES:
        # Check the cache first
        return origAppliance

    ioURL = "https://{webhost}/v2/{pathName}/manifests/{tag}" "".format(
        webhost=registryName, pathName=imageName, tag=tag
    )
    response = web_session.head(ioURL)
    if not response.ok:
        raise ApplianceImageNotFound(origAppliance, ioURL, response.status_code)
    else:
        KNOWN_EXTANT_IMAGES.add(origAppliance)
        return origAppliance


def requestCheckDockerIo(origAppliance: str, imageName: str, tag: str) -> bool:
    """
    Check docker.io to see if an image exists using the requests library.

    URL is based on the docker v2 schema.  Requires that an access token be fetched first.

    :param origAppliance: The full url of the docker image originally
        specified by the user (or the default). For example, ``ubuntu:latest``.
    :param imageName: The image, including path and excluding the tag. For example, ``ubuntu``.
    :param tag: The tag used at that docker image's registry. For example, ``latest``.
    :raises: ApplianceImageNotFound if no match is found.
    :return: Return True if match found.
    """
    if origAppliance in KNOWN_EXTANT_IMAGES:
        # Check the cache first
        return origAppliance

    # only official images like 'busybox' or 'ubuntu'
    if "/" not in imageName:
        imageName = "library/" + imageName

    token_url = "https://auth.docker.io/token?service=registry.docker.io&scope=repository:{repo}:pull".format(
        repo=imageName
    )
    requests_url = f"https://registry-1.docker.io/v2/{imageName}/manifests/{tag}"

    token = web_session.get(token_url)
    jsonToken = token.json()
    bearer = jsonToken["token"]
    response = web_session.head(
        requests_url, headers={"Authorization": f"Bearer {bearer}"}
    )
    if not response.ok:
        raise ApplianceImageNotFound(origAppliance, requests_url, response.status_code)
    else:
        KNOWN_EXTANT_IMAGES.add(origAppliance)
        return origAppliance


def logProcessContext(config: "Config") -> None:
    # toil.version.version (string) cannot be imported at top level because it conflicts with
    # toil.version (module) and Sphinx doesn't like that.
    from toil.version import version

    log.info("Running Toil version %s on host %s.", version, socket.gethostname())
    log.debug("Configuration: %s", config.__dict__)




import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp


def fn(job, i):
    job.log("i is: %s" % i, level=100)
    return i + 1


if __name__ == "__main__":
    jobstore: str = mkdtemp("tutorial_promises")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "INFO"
    options.clean = "always"

    j1 = Job.wrapJobFn(fn, 1)
    j2 = j1.addChildJobFn(fn, j1.rv())
    j3 = j1.addFollowOnJobFn(fn, j2.rv())

    with Toil(options) as toil:
        toil.start(j1)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# https://pytest.org/latest/example/pythoncollection.html

import json
import logging
from io import StringIO
from typing import Any, Dict, List, Optional, Tuple

from cwltest import utils
logger = logging.getLogger(__name__)

collect_ignore = ["spec"]


# Hook into Pytest for testing CWL conformance with Toil
# https://pytest.org/en/6.2.x/writing_plugins.html?highlight=conftest#conftest-py-local-per-directory-plugins
# See cwltool's reference implementation:
# https://github.com/common-workflow-language/cwltool/blob/05af6c1357c327b3146e9f5da40e7c0aa3e6d976/tests/cwl-conformance/cwltool-conftest.py
def pytest_cwl_execute_test(
        config: utils.CWLTestConfig,
        processfile: str,
        jobfile: Optional[str]
) -> Tuple[int, Optional[Dict[str, Any]]]:
    """Use Toil to execute CWL tests (equivalent to running toil-cwl-runner)."""
    from toil.cwl.cwltoil import main

    stdout = StringIO()
    argsl: List[str] = [f"--outdir={config.outdir}"]
    if config.runner_quiet:
        argsl.append("--quiet")
    elif config.verbose:
        argsl.append("--debug")
    argsl.extend(config.args)
    argsl.append(processfile)
    if jobfile:
        argsl.append(jobfile)
    try:
        result = main(args=argsl, stdout=stdout)
    except Exception as e:
        logger.error(e)
        return 1, {}
    out = stdout.getvalue()
    return result, json.loads(out) if out else {}



import textwrap
from queue import Queue

import pytest

import toil.batchSystems.gridengine
from toil.batchSystems.abstractGridEngineBatchSystem import ExceededRetryAttempts
from toil.common import Config
from toil.lib.misc import CalledProcessErrorStderr
from toil.test import ToilTest


class FakeBatchSystem:
    """
    Class that implements a minimal Batch System, needed to create a Worker (see below).
    """

    def __init__(self):
        self.config = self.__fake_config()

    def getWaitDuration(self):
        return 10

    def __fake_config(self):
        """
        Returns a dummy config for the batch system tests.  We need a workflowID to be set up
        since we are running tests without setting up a jobstore. This is the class version
        to be used when an instance is not available.

        :rtype: toil.common.Config
        """
        config = Config()
        from uuid import uuid4

        config.workflowID = str(uuid4())
        config.cleanWorkDir = "always"
        return config

    def with_retries(self, operation, *args, **kwargs):
        """
        The grid engine batch system needs a with_retries function when running the GridEngineThread, so fake one
        """
        return operation(*args, **kwargs)


def call_qstat_or_qacct(args, **_):
    # example outputs taken from https://2021.help.altair.com/2021.1/AltairGridEngine/8.7.0/UsersGuideGE.pdf
    qacct_info = {}
    job_id_info = {
        1: {"failed": True, "exit_code": 0, "completed": True},
        2: {"failed": True, "exit_code": 2, "completed": True},
        3: {"failed": False, "exit_code": 0, "completed": True},
        4: {"failed": False, "exit_code": 10, "completed": True},
        5: {"failed": False, "exit_code": 0, "completed": False},
    }
    for job_id, status_info in job_id_info.items():
        failed = 1 if status_info["failed"] else 0
        exit_status = status_info["exit_code"]
        qacct_info[job_id] = textwrap.dedent(
            f"""\
            ==============================================================
            qname all.q
            hostname kailua
            group users
            owner jondoe
            project NONE
            department defaultdepartment
            jobname Sleeper
            jobnumber 10
            taskid undefined
            account sge
            priority 0
            qsub_time Thu Mar 10 19:58:35 2011
            start_time Thu Mar 10 19:58:42 2011
            end_time Thu Mar 10 19:59:43 2011
            granted_pe NONE
            slots 1
            failed {failed}
            exit_status {exit_status}
            ru_wallclock 61
            ru_utime 0.070
            ru_stime 0.050
            ru_maxrss 1220
            ru_ixrss 0
            ru_ismrss 0
            ru_idrss 0
        """
        )
    if args[0] == "qstat":
        # This is guess for what qstat will return given a job. I'm unable to find an example for qstat.
        # This also assumes the second argument args[1] is -j, as that is what we try to use
        job_id = int(args[2])
        if job_id not in job_id_info.keys() or job_id_info[job_id]["completed"]:
            stderr = f"Following jobs do not exist {job_id}"
        else:
            # This is not the output of qstat when the job is running, and is just a guess
            # We test on the existence of the string "Following jobs do not exist", so this should be okay for now
            stderr = f"Job exists {job_id}"
        raise CalledProcessErrorStderr(2, args, stderr=stderr)
    elif args[0] == "qacct":
        if args[1] != "-j":
            # Documentation for qacct says if -j is not found then all jobs are listed
            # https://gridscheduler.sourceforge.net/htmlman/htmlman1/qacct.html
            # This is a guess for the output of qacct. We don't have a SGE cluster and I can't find a bare qacct example output online
            qacct_response = "\n".join(qacct_info.values())
        else:
            job_id = int(args[2])
            if job_id not in job_id_info.keys():
                # This is a guess of the behavior when the job does not exist. Since the behavior is unknown, this is not currently tested
                return ""
            qacct_response = qacct_info[job_id]

        return qacct_response


class GridEngineTest(ToilTest):
    """
    Class for unit-testing GridEngineBatchSystem
    """

    def setUp(self):
        self.monkeypatch = pytest.MonkeyPatch()
        self.worker = (
            toil.batchSystems.gridengine.GridEngineBatchSystem.GridEngineThread(
                newJobsQueue=Queue(),
                updatedJobsQueue=Queue(),
                killQueue=Queue(),
                killedJobsQueue=Queue(),
                boss=FakeBatchSystem(),
            )
        )

    ###
    ### Tests for coalesce_job_exit_codes for gridengine.
    ###

    def test_coalesce_job_exit_codes_one_exists(self):
        self.monkeypatch.setattr(
            toil.batchSystems.gridengine, "call_command", call_qstat_or_qacct
        )
        job_ids = ["1"]  # FAILED
        expected_result = [1]
        result = self.worker.coalesce_job_exit_codes(job_ids)
        assert result == expected_result, f"{result} != {expected_result}"

    def test_coalesce_job_exit_codes_one_still_running(self):
        self.monkeypatch.setattr(
            toil.batchSystems.gridengine, "call_command", call_qstat_or_qacct
        )
        job_ids = [
            "5"
        ]  # Still running. We currently raise an exception when this happens
        try:
            self.worker.coalesce_job_exit_codes(job_ids)
        except ExceededRetryAttempts:
            pass
        else:
            raise RuntimeError("Test did not raise an exception!")

    def test_coalesce_job_exit_codes_many_all_exist(self):
        self.monkeypatch.setattr(
            toil.batchSystems.gridengine, "call_command", call_qstat_or_qacct
        )
        job_ids = [
            "1",  # FAILED,
            "2",  # FAILED (with exit code that we ignore),
            "3",  # SUCCEEDED,
            "4",
        ]  # EXIT CODE 10
        # RUNNING and PENDING jobs should return None
        expected_result = [1, 1, 0, 10]
        result = self.worker.coalesce_job_exit_codes(job_ids)
        assert result == expected_result, f"{result} != {expected_result}"



import os
from argparse import ArgumentParser

from configargparse import SUPPRESS

from toil.version import baseVersion


def add_cwl_options(parser: ArgumentParser, suppress: bool = True) -> None:
    """
    Add CWL options to the parser. This only adds nonpositional CWL arguments.

    :param parser: Parser to add options to
    :param suppress: Suppress help output
    :return: None
    """
    suppress_help = SUPPRESS if suppress else None

    # These are options that we have to match cwltool
    # TODO: Are there still any Toil-specific options in here?
    parser.add_argument("--not-strict", action="store_true", help=suppress_help)
    parser.add_argument(
        "--enable-dev",
        action="store_true",
        help=suppress_help
        or suppress_help
        or "Enable loading and running development versions of CWL",
    )
    parser.add_argument(
        "--enable-ext",
        action="store_true",
        help=suppress_help
        or "Enable loading and running 'cwltool:' extensions to the CWL standards.",
        default=False,
    )
    parser.add_argument(
        "--quiet", dest="quiet", action="store_true", default=False, help=suppress_help
    )
    parser.add_argument(
        "--basedir", type=str, help=suppress_help
    )  # TODO: Might be hard-coded?
    parser.add_argument("--outdir", type=str, default=None, help=suppress_help)
    parser.add_argument(
        "--version",
        action="version",
        version=baseVersion,
        help=suppress_help or "show program's version number and exit",
    )
    parser.add_argument(
        "--log-dir",
        type=str,
        default="",
        help=suppress_help
        or "Log your tools stdout/stderr to this location outside of container",
    )
    # this is as a result of suppressed help statements not working well with mutually_exclusive_groups, which will
    # cause an assertion error
    # https://github.com/python/cpython/issues/62090
    dockergroup = (
        parser.add_mutually_exclusive_group()
        if not suppress_help
        else parser.add_argument_group()
    )
    dockergroup.add_argument(
        "--user-space-docker-cmd",
        help=suppress_help
        or "(Linux/OS X only) Specify a user space docker command (like "
        "udocker or dx-docker) that will be used to call 'pull' and 'run'",
    )
    dockergroup.add_argument(
        "--singularity",
        action="store_true",
        default=False,
        help=suppress_help
        or "Use Singularity runtime for running containers. "
        "Requires Singularity v2.6.1+ and Linux with kernel version v3.18+ or "
        "with overlayfs support backported.",
    )
    dockergroup.add_argument(
        "--podman",
        action="store_true",
        default=False,
        help=suppress_help or "Use Podman runtime for running containers. ",
    )
    dockergroup.add_argument(
        "--no-container",
        action="store_true",
        help=suppress_help
        or "Do not execute jobs in a "
        "Docker container, even when `DockerRequirement` "
        "is specified under `hints`.",
    )
    dockergroup.add_argument(
        "--leave-container",
        action="store_false",
        default=True,
        help=suppress_help
        or "Do not delete Docker container used by jobs after they exit",
        dest="rm_container",
    )
    parser.add_argument(
        "--custom-net",
        help=suppress_help
        or "Specify docker network name to pass to docker run command",
    )
    cidgroup = parser.add_argument_group(
        "Options for recording the Docker container identifier into a file."
    )
    cidgroup.add_argument(
        # Disabled as containerid is now saved by default
        "--record-container-id",
        action="store_true",
        default=False,
        help=suppress_help or SUPPRESS,
        dest="record_container_id",
    )

    cidgroup.add_argument(
        "--cidfile-dir",
        type=str,
        help=suppress_help
        or "Store the Docker container ID into a file in the specified directory.",
        default=None,
        dest="cidfile_dir",
    )

    cidgroup.add_argument(
        "--cidfile-prefix",
        type=str,
        help=suppress_help
        or "Specify a prefix to the container ID filename. "
        "Final file name will be followed by a timestamp. "
        "The default is no prefix.",
        default=None,
        dest="cidfile_prefix",
    )

    parser.add_argument(
        "--no-prepull",
        action="store_true",
        default=False,
        help=suppress_help
        or "Do not prepull the container prior to running the workflow",
    )

    parser.add_argument(
        "--preserve-environment",
        type=str,
        nargs="+",
        help=suppress_help
        or "Preserve specified environment variables when running" " CommandLineTools",
        metavar=("VAR1 VAR2"),
        default=("PATH",),
        dest="preserve_environment",
    )
    parser.add_argument(
        "--preserve-entire-environment",
        action="store_true",
        help=suppress_help
        or "Preserve all environment variable when running CommandLineTools.",
        default=False,
        dest="preserve_entire_environment",
    )
    parser.add_argument(
        "--beta-dependency-resolvers-configuration", default=None, help=suppress_help
    )
    parser.add_argument(
        "--beta-dependencies-directory", default=None, help=suppress_help
    )
    parser.add_argument(
        "--beta-use-biocontainers",
        default=None,
        action="store_true",
        help=suppress_help,
    )
    parser.add_argument(
        "--beta-conda-dependencies",
        default=None,
        action="store_true",
        help=suppress_help,
    )
    parser.add_argument(
        "--tmpdir-prefix",
        type=str,
        help=suppress_help or "Path prefix for temporary directories",
        default=None,
    )
    parser.add_argument(
        "--tmp-outdir-prefix",
        type=str,
        help=suppress_help or "Path prefix for intermediate output directories",
        default=None,
    )
    parser.add_argument(
        "--force-docker-pull",
        action="store_true",
        default=False,
        dest="force_docker_pull",
        help=suppress_help or "Pull latest docker image even if it is locally present",
    )
    parser.add_argument(
        "--no-match-user",
        action="store_true",
        default=False,
        help=suppress_help or "Disable passing the current uid to `docker run --user`",
    )
    parser.add_argument(
        "--no-read-only",
        action="store_true",
        default=False,
        help=suppress_help or "Do not set root directory in the container as read-only",
    )
    parser.add_argument(
        "--strict-memory-limit",
        action="store_true",
        help=suppress_help
        or "When running with "
        "software containers and the Docker engine, pass either the "
        "calculated memory allocation from ResourceRequirements or the "
        "default of 1 gigabyte to Docker's --memory option.",
    )
    parser.add_argument(
        "--strict-cpu-limit",
        action="store_true",
        help=suppress_help
        or "When running with "
        "software containers and the Docker engine, pass either the "
        "calculated cpu allocation from ResourceRequirements or the "
        "default of 1 core to Docker's --cpu option. "
        "Requires docker version >= v1.13.",
    )
    parser.add_argument(
        "--relax-path-checks",
        action="store_true",
        default=False,
        help=suppress_help
        or "Relax requirements on path names to permit " "spaces and hash characters.",
        dest="relax_path_checks",
    )
    parser.add_argument(
        "--default-container",
        help=suppress_help
        or "Specify a default docker container that will be "
        "used if the workflow fails to specify one.",
    )
    parser.add_argument(
        "--disable-validate",
        dest="do_validate",
        action="store_false",
        default=True,
        help=suppress_help or SUPPRESS,
    )
    parser.add_argument(
        "--fast-parser",
        dest="fast_parser",
        action="store_true",
        default=False,
        help=suppress_help or SUPPRESS,
    )
    # same workaround as dockergroup
    checkgroup = (
        parser.add_mutually_exclusive_group()
        if not suppress_help
        else parser.add_argument_group()
    )
    checkgroup.add_argument(
        "--compute-checksum",
        action="store_true",
        default=True,
        help=suppress_help or "Compute checksum of contents while collecting outputs",
        dest="compute_checksum",
    )
    checkgroup.add_argument(
        "--no-compute-checksum",
        action="store_false",
        help=suppress_help
        or "Do not compute checksum of contents while collecting outputs",
        dest="compute_checksum",
    )

    parser.add_argument(
        "--eval-timeout",
        help=suppress_help
        or "Time to wait for a Javascript expression to evaluate before giving "
        "an error, default 20s.",
        type=float,
        default=20,
    )
    parser.add_argument(
        "--overrides",
        type=str,
        default=None,
        help=suppress_help or "Read process requirement overrides from file.",
    )

    parser.add_argument(
        "--mpi-config-file",
        type=str,
        default=None,
        help=suppress_help
        or "Platform specific configuration for MPI (parallel "
        "launcher, its flag etc). See the cwltool README "
        "section 'Running MPI-based tools' for details of the format: "
        "https://github.com/common-workflow-language/cwltool#running-mpi-based-tools-that-need-to-be-launched",
    )

    provgroup = parser.add_argument_group(
        "Options for recording provenance information of the execution"
    )
    provgroup.add_argument(
        "--provenance",
        help=suppress_help
        or "Save provenance to specified folder as a "
        "Research Object that captures and aggregates "
        "workflow execution and data products.",
        type=str,
    )

    provgroup.add_argument(
        "--enable-user-provenance",
        default=False,
        action="store_true",
        help=suppress_help or "Record user account info as part of provenance.",
        dest="user_provenance",
    )
    provgroup.add_argument(
        "--disable-user-provenance",
        default=False,
        action="store_false",
        help=suppress_help or "Do not record user account info in provenance.",
        dest="user_provenance",
    )
    provgroup.add_argument(
        "--enable-host-provenance",
        default=False,
        action="store_true",
        help=suppress_help or "Record host info as part of provenance.",
        dest="host_provenance",
    )
    provgroup.add_argument(
        "--disable-host-provenance",
        default=False,
        action="store_false",
        help=suppress_help or "Do not record host info in provenance.",
        dest="host_provenance",
    )
    provgroup.add_argument(
        "--orcid",
        help=suppress_help
        or "Record user ORCID identifier as part of "
        "provenance, e.g. https://orcid.org/0000-0002-1825-0097 "
        "or 0000-0002-1825-0097. Alternatively the environment variable "
        "ORCID may be set.",
        dest="orcid",
        default=os.environ.get("ORCID", ""),
        type=str,
    )
    provgroup.add_argument(
        "--full-name",
        help=suppress_help
        or "Record full name of user as part of provenance, "
        "e.g. Josiah Carberry. You may need to use shell quotes to preserve "
        "spaces. Alternatively the environment variable CWL_FULL_NAME may "
        "be set.",
        dest="cwl_full_name",
        default=os.environ.get("CWL_FULL_NAME", ""),
        type=str,
    )

    # These are Toil-specific options
    parser.add_argument(
        "--bypass-file-store",
        action="store_true",
        default=False,
        help=suppress_help
        or "Do not use Toil's file store and assume all "
        "paths are accessible in place from all nodes.",
        dest="bypass_file_store",
    )
    parser.add_argument(
        "--reference-inputs",
        action="store_true",
        default=False,
        help=suppress_help
        or "Do not copy remote inputs into Toil's file "
        "store and assume they are accessible in place from "
        "all nodes.",
        dest="reference_inputs",
    )
    parser.add_argument(
        "--disable-streaming",
        action="store_true",
        default=False,
        help=suppress_help
        or "Disable file streaming for files that have 'streamable' flag True.",
        dest="disable_streaming",
    )
    ram_group = (
        parser.add_mutually_exclusive_group()
        if not suppress_help
        else parser.add_argument_group()
    )
    ram_group.add_argument(
        "--cwl-default-ram",
        action="store_true",
        default=True,
        help=suppress_help or "Apply CWL specification default ramMin.",
        dest="cwl_default_ram",
    )
    ram_group.add_argument(
        "--no-cwl-default-ram",
        action="store_false",
        help=suppress_help
        or "Do not apply CWL specification default ramMin, so that Toil --defaultMemory applies.",
        dest="cwl_default_ram",
    )
    parser.add_argument(
        "--destBucket",
        type=str,
        help=suppress_help or "Specify a cloud bucket endpoint for output files.",
    )
    parser.add_argument(
        "--cachedir",
        type=str,
        help=suppress_help
        or "Directory to cache intermediate workflow outputs to avoid "
        "recomputing steps. Can be very helpful in the development and "
        "troubleshooting of CWL documents. This automatically bypasses the file store."
        " Not to be confused with --caching.",
        dest="cachedir"
    )



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp


class HelloWorld(Job):
    def __init__(self, message):
        Job.__init__(self)
        self.message = message

    def run(self, fileStore):
        return f"Hello, world!, here's a message: {self.message}"


if __name__ == "__main__":
    jobstore: str = mkdtemp("tutorial_invokeworkflow")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "OFF"
    options.clean = "always"

    hello_job = HelloWorld("Woot")

    with Toil(options) as toil:
        print(toil.start(hello_job))
